r"""
The submodule in `cl_algorithms` for [EWC (Elastic Weight Consolidation) algorithm](https://www.pnas.org/doi/10.1073/pnas.1611835114).
"""

__all__ = ["EWC"]

import logging
from copy import deepcopy
from typing import Any

import torch
from torch import Tensor, nn

from clarena.backbones import CLBackbone
from clarena.cl_algorithms import Finetuning
from clarena.cl_algorithms.regularisers import ParameterChangeReg
from clarena.cl_heads import HeadsCIL, HeadsTIL

# always get logger for built-in logging in each module
pylogger = logging.getLogger(__name__)


class EWC(Finetuning):
    r"""EWC (Elastic Weight Consolidation) algorithm.

    [EWC (Elastic Weight Consolidation, 2017)](https://www.pnas.org/doi/10.1073/pnas.1611835114) is a regularisation-based continual learning approach that calculates parameter importance for the previous tasks and penalises the current task loss with the importance of the parameters.

    We implement EWC as a subclass of Finetuning algorithm, as EWC has the same `forward()`, `validation_step()` and `test_step()` method as `Finetuning` class.
    """

    def __init__(
        self,
        backbone: CLBackbone,
        heads: HeadsTIL | HeadsCIL,
        parameter_change_reg_factor: float,
        parameter_change_reg_p_norm: float,
    ) -> None:
        r"""Initialise the HAT algorithm with the network.

        **Args:**
        - **backbone** (`CLBackbone`): backbone network.
        - **heads** (`HeadsTIL` | `HeadsCIL`): output heads.
        - **parameter_change_reg_factor** (`float`): the parameter change regularisation factor. It controls the strength of preventing forgetting.
        - **parameter_change_reg_p_norm** (`float`): the norm of the distance of parameters between previous tasks and current task in the parameter change regularisation.

        """
        Finetuning.__init__(self, backbone=backbone, heads=heads)

        self.parameter_importance: dict[str, dict[str, Tensor]] = {}
        r"""Store the parameter importance of each previous task. Keys are task IDs (string type) and values are the corresponding importance. Each importance entity is a dict where keys are parameter names (named by `named_parameters()` of the `nn.Module`) and values are the importance tensor for the layer. It has the same shape as the parameters of the layer.
        """
        self.previous_task_backbones: dict[str, nn.Module] = {}
        r"""Store the backbone models of the previous tasks. Keys are task IDs (string type) and values are the corresponding models. Each model is a `nn.Module` backbone after the corresponding previous task was trained.
        
        Some would argue that since we could store the model of the previous tasks, why don't we test the task directly with the stored model, instead of doing the less easier EWC thing? The thing is, EWC only uses the model of the previous tasks to train current and future tasks, which aggregate them into a single model. Once the training of the task is done, the storage for those parameters can be released. However, this make the future tasks not able to use EWC anymore, which is a disadvantage for EWC.
        """

        self.parameter_change_reg_factor = parameter_change_reg_factor
        r"""Store parameter change regularisation factor."""
        self.parameter_change_reg_p_norm = parameter_change_reg_p_norm
        r"""Store norm of the distance used in parameter change regularisation."""
        self.parameter_change_reg = ParameterChangeReg(
            factor=parameter_change_reg_factor,
            p_norm=parameter_change_reg_p_norm,
        )
        r"""Initialise and store the parameter change regulariser."""

        EWC.sanity_check(self)

    def sanity_check(self) -> None:
        r"""Check the sanity of the arguments.

        **Raises:**
        - **ValueError**: If the regularisation factor is not positive.
        """

        if self.parameter_change_reg_factor <= 0:
            raise ValueError(
                f"The parameter change regularisation factor should be positive, but got {self.parameter_change_reg_factor}."
            )

    def calculate_parameter_importance(self) -> None:
        r"""Calculate the parameter importance for the learned task. This is only called after the training of a task, which is the last previous task $t-1$. The calculated importance is stored in `self.parameter_importance[self.task_id]` for constructing the regularisation loss in the future tasks.

        According to [the EWC paper](https://www.pnas.org/doi/10.1073/pnas.1611835114), the importance tensor is a Laplace approximation to Fisher information matrix by taking the digonal, i.e. $F_i$, where $i$ is the index of a parameter. The calculation is not following that theory but the derived formula below:

        $$\omega_i = F_i  =\frac{1}{N_{t-1}} \sum_{(\mathbf{x}, y)\in \mathcal{D}^{(t-1)}_{\text{train}}} \left[\frac{\partial l(f^{(t-1)}\left(\mathbf{x}, \theta), y\right)}{\partial \theta_i}\right]^2$$

        For a parameter $i$, its importance is the magnitude (square here) of gradient of the loss of model just trained over the training data just used. The $l$ is the classification loss. It shows the sensitivity of the loss to the parameter. The larger it is, the more it changed the performance (which is the loss) of the model, which indicates the importance of the parameter.
        """
        parameter_importance_t = {}

        # set model to evaluation mode to prevent updating the model parameters
        self.eval()

        # get the training data
        last_task_train_dataloaders = self.trainer.datamodule.train_dataloader()

        # initialise the accumulation of the squared gradients
        num_data = 0
        for param_name, param in self.backbone.named_parameters():
            parameter_importance_t[param_name] = torch.zeros_like(param)

        for x, y in last_task_train_dataloaders:

            # move data to device manually
            x = x.to(self.device)
            y = y.to(self.device)

            batch_size = len(y)
            num_data += batch_size

            # compute the gradients within a batch
            self.backbone.zero_grad()  # reset gradients
            logits, _ = self.forward(x, stage="train", task_id=self.task_id)
            loss_cls = self.criterion(logits, y)
            loss_cls.backward()  # compute gradients

            # collect and accumulate the squared gradients into parameter importance
            for param_name, param in self.backbone.named_parameters():
                parameter_importance_t[param_name] += batch_size * param.grad**2

        num_params = sum(p.numel() for p in self.backbone.parameters())

        for param_name, param in self.backbone.named_parameters():
            parameter_importance_t[param_name] /= (
                num_data * num_params
            )  # average over data and parameters

        self.parameter_importance[self.task_id] = parameter_importance_t

    def training_step(self, batch: Any) -> dict[str, Tensor]:
        r"""Training step for current task `self.task_id`.

        **Args:**
        - **batch** (`Any`): a batch of training data.

        **Returns:**
        - **outputs** (`dict[str, Tensor]`): a dictionary contains loss and other metrics from this training step. Key (`str`) is the metrics name, value (`Tensor`) is the metrics. Must include the key 'loss' which is total loss in the case of automatic optimization, according to PyTorch Lightning docs.
        """
        x, y = batch

        # classification loss
        logits, hidden_features = self.forward(x, stage="train", task_id=self.task_id)
        loss_cls = self.criterion(logits, y)

        # regularisation loss. See equation (3) in the [EWC paper](https://www.pnas.org/doi/10.1073/pnas.1611835114).
        loss_reg = 0.0
        for previous_task_id in range(1, self.task_id):
            # sum over all previous models, because [EWC paper](https://www.pnas.org/doi/10.1073/pnas.1611835114) says: "When moving to a third task, task C, EWC will try to keep the network parameters close to the learned parameters of both tasks A and B. This can be enforced either with two separate penalties or as one by noting that the sum of two quadratic penalties is itself a quadratic penalty."
            loss_reg += self.parameter_change_reg(
                target_model=self.backbone,
                ref_model=self.previous_task_backbones[previous_task_id],
                weights=self.parameter_importance[previous_task_id],
            )

        if self.task_id != 1:
            loss_reg /= (
                self.task_id
            )  # average over tasks to avoid linear increase of the regularisation loss

        # total loss
        loss = loss_cls + loss_reg

        # accuracy of the batch
        acc = (logits.argmax(dim=1) == y).float().mean()

        return {
            "loss": loss,  # Return loss is essential for training step, or backpropagation will fail
            "loss_cls": loss_cls,
            "loss_reg": loss_reg,
            "acc": acc,
            "hidden_features": hidden_features,
        }

    def on_train_end(self) -> None:
        r"""Calculate the parameter importance and store the backbone model after the training of a task.

        The calculated importance and model are stored in `self.parameter_importance[self.task_id]` and `self.previous_task_backbones[self.task_id]` respectively for constructing the regularisation loss in the future tasks.
        """
        self.calculate_parameter_importance()

        previous_backbone = deepcopy(self.backbone)
        previous_backbone.eval()  # set the store model to evaluation mode to prevent updating
        self.previous_task_backbones[self.task_id] = previous_backbone



r"""
The submodule in `cl_algorithms` for Finetuning algorithm.
"""

__all__ = ["Finetuning"]

import logging
from typing import Any

import torch
from torch import Tensor
from torch.utils.data import DataLoader

from clarena.backbones import CLBackbone
from clarena.cl_algorithms import CLAlgorithm
from clarena.cl_heads import HeadsCIL, HeadsTIL

# always get logger for built-in logging in each module
pylogger = logging.getLogger(__name__)


class Finetuning(CLAlgorithm):
    r"""Finetuning algorithm.

    It is the most naive way for task-incremental learning. It simply initialises the backbone from the last task when training new task.
    """

    def __init__(
        self,
        backbone: CLBackbone,
        heads: HeadsTIL | HeadsCIL,
    ) -> None:
        r"""Initialise the Finetuning algorithm with the network. It has no additional hyperparamaters.

        **Args:**
        - **backbone** (`CLBackbone`): backbone network.
        - **heads** (`HeadsTIL` | `HeadsCIL`): output heads.
        """
        CLAlgorithm.__init__(self, backbone=backbone, heads=heads)

    def forward(self, input: Tensor, stage: str, task_id: int | None = None) -> Tensor:
        r"""The forward pass for data from task `task_id`. Note that it is nothing to do with `forward()` method in `nn.Module`.

        **Args:**
        - **input** (`Tensor`): The input tensor from data.
        - **stage** (`str`): the stage of the forward pass, should be one of the following:
            1. 'train': training stage.
            2. 'validation': validation stage.
            3. 'test': testing stage.
        - **task_id** (`int`): the task ID where the data are from. If stage is 'train' or `validation`, it is usually from the current task `self.task_id`. If stage is 'test', it could be from any seen task. In TIL, the task IDs of test data are provided thus this argument can be used. In CIL, they are not provided, so it is just a placeholder for API consistence but never used, and best practices are not to provide this argument and leave it as the default value. Finetuning algorithm works both for TIL and CIL.

        **Returns:**
        - **logits** (`Tensor`): the output logits tensor.
        - **hidden_features** (`dict[str, Tensor]`): the hidden features (after activation) in each weighted layer. Key (`str`) is the weighted layer name, value (`Tensor`) is the hidden feature tensor. This is used for the continual learning algorithms that need to use the hidden features for various purposes. Although Finetuning algorithm does not need this, it is still provided for API consistence for other algorithms inherited this `forward()` method of `Finetuning` class.
        """
        feature, hidden_features = self.backbone(input, stage=stage, task_id=task_id)
        logits = self.heads(feature, task_id)
        return logits, hidden_features

    def training_step(self, batch: Any) -> dict[str, Tensor]:
        """Training step for current task `self.task_id`.

        **Args:**
        - **batch** (`Any`): a batch of training data.

        **Returns:**
        - **outputs** (`dict[str, Tensor]`): a dictionary contains loss and other metrics from this training step. Key (`str`) is the metrics name, value (`Tensor`) is the metrics. Must include the key 'loss' which is total loss in the case of automatic optimization, according to PyTorch Lightning docs.
        """
        x, y = batch

        # classification loss
        logits, hidden_features = self.forward(x, stage="train", task_id=self.task_id)
        loss_cls = self.criterion(logits, y)

        # total loss
        loss = loss_cls

        # accuracy of the batch
        acc = (logits.argmax(dim=1) == y).float().mean()

        return {
            "loss": loss,  # Return loss is essential for training step, or backpropagation will fail
            "loss_cls": loss_cls,
            "acc": acc,  # Return other metrics for lightning loggers callback to handle at `on_train_batch_end()`
            "hidden_features": hidden_features,
        }

    def validation_step(self, batch: Any) -> dict[str, Tensor]:
        r"""Validation step for current task `self.task_id`.

        **Args:**
        - **batch** (`Any`): a batch of validation data.

        **Returns:**
        - **outputs** (`dict[str, Tensor]`): a dictionary contains loss and other metrics from this validation step. Key (`str`) is the metrics name, value (`Tensor`) is the metrics.
        """
        x, y = batch
        logits, hidden_features = self.forward(
            x,
            stage="validation",
            task_id=self.task_id,
        )
        loss_cls = self.criterion(logits, y)
        acc = (logits.argmax(dim=1) == y).float().mean()

        # Return metrics for lightning loggers callback to handle at `on_validation_batch_end()`
        return {
            "loss_cls": loss_cls,
            "acc": acc,
        }

    def test_step(
        self, batch: DataLoader, batch_idx: int, dataloader_idx: int = 0
    ) -> dict[str, Tensor]:
        r"""Test step for current task `self.task_id`, which tests for all seen tasks indexed by `dataloader_idx`.

        **Args:**
        - **batch** (`Any`): a batch of test data.
        - **dataloader_idx** (`int`): the task ID of seen tasks to be tested. A default value of 0 is given otherwise the LightningModule will raise a `RuntimeError`.

        **Returns:**
        - **outputs** (`dict[str, Tensor]`): a dictionary contains loss and other metrics from this test step. Key (`str`) is the metrics name, value (`Tensor`) is the metrics.
        """
        test_task_id = dataloader_idx + 1

        x, y = batch
        logits, hidden_features = self.forward(
            x, stage="test", task_id=test_task_id
        )  # use the corresponding head to test (instead of the current task `self.task_id`)
        loss_cls = self.criterion(logits, y)
        acc = (logits.argmax(dim=1) == y).float().mean()

        # Return metrics for lightning loggers callback to handle at `on_test_batch_end()`
        return {
            "loss_cls": loss_cls,
            "acc": acc,
        }



"""The submodule in `utils` for saving data in the experiment as files."""

__all__ = [
    "update_test_acc_to_csv",
    "update_test_loss_cls_to_csv",
]


import csv
import os

from torchmetrics import MeanMetric

from clarena.utils import MeanMetricBatch


def update_test_acc_to_csv(
    test_acc_metric: dict[int, MeanMetricBatch], csv_path: str
) -> None:
    """Update the test accuracy metrics of task 1 to `num_tasks` at the last line to an existing csv file. A new file will be created if not existing. Used in `CLMetricsCallback`.

    **Args:**
    - **test_acc_metric** (`dict[int, MeanMetricBatch]`): classification accuracy of the test data of each seen task. Accumulated and calculated from the test batches.
    - **csv_path** (`str`): save the test metric to path. E.g. './outputs/expr_name/1970-01-01_00-00-00/acc.csv'.
    """
    num_tasks = len(test_acc_metric)
    fieldnames = ["after_training_task", "average_accuracy"] + [
        f"test_on_task_{task_id}" for task_id in range(1, num_tasks + 1)
    ]

    new_line = {"after_training_task": num_tasks}  # construct the first column

    # construct the columns and calculate the average accuracy over tasks at the same time
    average_accuracy_over_tasks = MeanMetric()
    for task_id in range(1, num_tasks + 1):
        acc = test_acc_metric[f"{task_id}"].compute().item()
        new_line[f"test_on_task_{task_id}"] = acc
        average_accuracy_over_tasks(acc)
    new_line["average_accuracy"] = average_accuracy_over_tasks.compute().item()

    # write to the csv file
    is_first = not os.path.exists(csv_path)
    if not is_first:
        with open(csv_path, "r", encoding="utf-8") as file:
            lines = file.readlines()
            del lines[0]
    # write header
    with open(csv_path, "w", encoding="utf-8") as file:
        writer = csv.DictWriter(file, fieldnames=fieldnames)
        writer.writeheader()
    # write metrics
    with open(csv_path, "a", encoding="utf-8") as file:
        if not is_first:
            file.writelines(lines)
        writer = csv.DictWriter(file, fieldnames=fieldnames)
        writer.writerow(new_line)


def update_test_loss_cls_to_csv(
    test_loss_cls_metric: dict[int, MeanMetricBatch], csv_path: str
) -> None:
    """Update the test classification loss metrics of task 1 to `num_tasks` at the last line to an existing csv file. A new file will be created if not existing. Used in `CLMetricsCallback`.

    **Args:**
    - **test_loss_cls_metric** (`dict[int, MeanMetricBatch]`): classification loss of the test data of each seen task. Accumulated and calculated from the test batches.
    - **csv_path** (`str`): save the test metric to path. E.g. './outputs/expr_name/1970-01-01_00-00-00/loss_cls.csv'.
    """
    num_tasks = len(test_loss_cls_metric)
    fieldnames = ["after_training_task", "average_classification_loss"] + [
        f"test_on_task_{task_id}" for task_id in range(1, num_tasks + 1)
    ]

    new_line = {"after_training_task": num_tasks}  # construct the first column

    # write to the columns and calculate the average classification loss over tasks at the same time
    average_classification_loss_over_tasks = MeanMetric()
    for task_id in range(1, num_tasks + 1):
        # task_id = dataloader_idx
        loss_cls = test_loss_cls_metric[f"{task_id}"].compute().item()
        new_line[f"test_on_task_{task_id}"] = loss_cls
        average_classification_loss_over_tasks(loss_cls)
    new_line["average_classification_loss"] = (
        average_classification_loss_over_tasks.compute().item()
    )

    # write to the csv file
    is_first = not os.path.exists(csv_path)
    if not is_first:
        with open(csv_path, "r", encoding="utf-8") as file:
            lines = file.readlines()
            del lines[0]
    # write header
    with open(csv_path, "w", encoding="utf-8") as file:
        writer = csv.DictWriter(file, fieldnames=fieldnames)
        writer.writeheader()
    # write metrics
    with open(csv_path, "a", encoding="utf-8") as file:
        if not is_first:
            file.writelines(lines)
        writer = csv.DictWriter(file, fieldnames=fieldnames)
        writer.writerow(new_line)



r"""The submodule in `cl_heads` for CIL heads."""

__all__ = ["HeadsCIL"]

import torch
from torch import Tensor, nn


class HeadsCIL(nn.Module):
    r"""The output heads for Class-Incremental Learning (CIL). Head of all classes from CIL tasks takes the output from backbone network and forwards it into logits for predicting classes of all tasks."""

    def __init__(self, input_dim: int) -> None:
        """Initializes a CIL heads object with no heads.

        **Args:**
        - **input_dim** (`int`): the input dimension of the heads. Must be equal to the `output_dim` of the connected backbone.
        """
        nn.Module.__init__(self)

        self.heads: nn.ModuleDict = nn.ModuleDict()  # initially no heads
        """CIL output heads are stored in a `ModuleDict`. Keys are task IDs (string type) and values are the corresponding `nn.Linear` heads. We use `ModuleDict` rather than `dict` to make sure `LightningModule` can track these model parameters for the purpose of, such as automatically to device, recorded in model summaries. """
        self.input_dim: int = input_dim
        """The input dimension of the heads. Used when creating new heads."""

        self.task_id: int
        """Task ID counter indicating which task is being processed. Self updated during the task loop."""

    def setup_task_id(self, task_id: int, num_classes_t: int) -> None:
        """Create the output head when task `task_id` arrives if there's no. This must be done before `forward()` is called.

        **Args:**
        - **task_id** (`int`): the target task ID.
        - **num_classes_t** (`int`): the number of classes in the task.
        """
        self.task_id = task_id
        if self.task_id not in self.heads.keys():
            self.heads[f"{self.task_id}"] = nn.Linear(self.input_dim, num_classes_t)

    def forward(self, feature: Tensor, task_id: int | None = None) -> Tensor:
        r"""The forward pass for data. The information of which `task_id` the data are from is not provided. The head for all classes is selected and the feature is passed.

        **Args:**
        - **feature** (`Tensor`): the feature tensor from the backbone network.
        - **task_id** (`int` or `None`): the task ID where the data are from. In CIL, it is just a placeholder for API consistence with the TIL heads but never used. Best practices are not to provide this argument and leave it as the default value.

        **Returns:**
        - **logits** (`Tensor`): the output logits tensor.
        """
        logits = torch.cat(
            [self.heads[f"{t}"](feature) for t in range(1, self.task_id + 1)], dim=-1
        )  # concatenate logits of classes from all heads

        return logits



r"""
The submodule in `cl_datasets` for Permuted MNIST dataset.
"""

__all__ = ["PermutedMNIST"]

import logging
from typing import Callable

import torch
from torch.utils.data import Dataset, random_split
from torchvision.datasets import MNIST
from torchvision.transforms import transforms

from clarena.cl_datasets import CLPermutedDataset

# always get logger for built-in logging in each module
pylogger = logging.getLogger(__name__)


class PermutedMNIST(CLPermutedDataset):
    r"""Permuted MNIST dataset. The [original MNIST dataset](http://yann.lecun.com/exdb/mnist/) is a collection of handwritten digits. It consists of 70,000 28x28 B&W images in 10 classes (correspond to 10 digits), with 7000 images per class. There are 60,000 training examples and 10,000 test examples."""

    num_classes: int = 10
    """The number of classes in MNIST."""

    img_size: torch.Size = torch.Size([1, 28, 28])
    """The size of MNIST images."""

    mean_original: tuple[float] = (0.1307,)
    """The mean values for normalisation."""

    std_original: tuple[float] = (0.3081,)
    """The standard deviatfion values for normalisation."""

    def __init__(
        self,
        root: str,
        num_tasks: int,
        validation_percentage: float,
        batch_size: int = 1,
        num_workers: int = 8,
        custom_transforms: Callable | transforms.Compose | None = None,
        custom_target_transforms: Callable | transforms.Compose | None = None,
        permutation_mode: str = "first_channel_only",
        permutation_seeds: list[int] | None = None,
    ) -> None:
        r"""Initialise the Permuted MNIST dataset.

        **Args:**
        - **root** (`str`): the root directory where the original MNIST data 'MNIST/raw/train-images-idx3-ubyte' and 'MNIST/raw/t10k-images-idx3-ubyte' live.
        - **num_tasks** (`int`): the maximum number of tasks supported by the CL dataset.
        - **validation_percentage** (`float`): the percentage to randomly split some of the training data into validation data.
        - **batch_size** (`int`): The batch size in train, val, test dataloader.
        - **num_workers** (`int`): the number of workers for dataloaders.
        - **custom_transforms** (`transform` or `transforms.Compose` or `None`): the custom transforms to apply to ONLY TRAIN dataset. Can be a single transform, composed transforms or no transform.
        `ToTensor()`, normalise, permute and so on are not included.
        - **custom_target_transforms** (`transform` or `transforms.Compose` or `None`): the custom target transforms to apply to dataset labels. Can be a single transform, composed transforms or no transform. CL class mapping is not included.
        - **permutation_mode** (`str`): the mode of permutation, should be one of the following:
            1. 'all': permute all pixels.
            2. 'by_channel': permute channel by channel separately. All channels are applied the same permutation order.
            3. 'first_channel_only': permute only the first channel.
        - **permutation_seeds** (`list[int]` or `None`): the seeds for permutation operations used to construct tasks. Make sure it has the same number of seeds as `num_tasks`. Default is None, which creates a list of seeds from 1 to `num_tasks`.
        """
        CLPermutedDataset.__init__(
            self,
            root=root,
            num_tasks=num_tasks,
            validation_percentage=validation_percentage,
            batch_size=batch_size,
            num_workers=num_workers,
            custom_transforms=custom_transforms,
            custom_target_transforms=custom_target_transforms,
            permutation_mode=permutation_mode,
            permutation_seeds=permutation_seeds,
        )

    def prepare_data(self) -> None:
        r"""Download the original MNIST dataset if haven't."""
        # just download
        MNIST(root=self.root, train=True, download=True)
        MNIST(root=self.root, train=False, download=True)

        pylogger.debug(
            "The original MNIST dataset has been downloaded to %s.", self.root
        )

    def train_and_val_dataset(self) -> tuple[Dataset, Dataset]:
        """Get the training and validation dataset of task `self.task_id`.

        **Returns:**
        - **train_and_val_dataset** (`tuple[Dataset, Dataset]`): the train and validation dataset of task `self.task_id`.
        """
        dataset_train_and_val = MNIST(
            root=self.root,
            train=True,
            transform=self.train_and_val_transforms(to_tensor=True),
            download=False,
        )
        return random_split(
            dataset_train_and_val,
            lengths=[1 - self.validation_percentage, self.validation_percentage],
        )

    def test_dataset(self) -> Dataset:
        r"""Get the test dataset of task `self.task_id`.

        **Returns:**
        - **test_dataset** (`Dataset`): the test dataset of task `self.task_id`.
        """

        return MNIST(
            root=self.root,
            train=False,
            transform=self.test_transforms(to_tensor=True),
            download=False,
        )



r"""
The submodule in `callbacks` for `HATMetricsCallback`.
"""

__all__ = ["HATMetricsCallback"]

import logging
import os
from typing import Any

from lightning import Callback, Trainer

from clarena.cl_algorithms import HAT, CLAlgorithm
from clarena.utils import plot

# always get logger for built-in logging in each module
pylogger = logging.getLogger(__name__)


class HATMetricsCallback(Callback):
    r"""Provides all actions that are related to metrics used for [HAT (Hard Attention to the Task)](http://proceedings.mlr.press/v80/serra18a) algorithm, which include:

    - Visualising mask and cumulative mask figures during training and testing as figures.
    - Logging network capacity during training. See the "Evaluation Metrics" section in chapter 4.1 in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9) for more details about network capacity.

    Lightning provides `self.log()` to log metrics in `LightningModule` where our `CLAlgorithm` based. You can put `self.log()` here if you don't want to mess up the `CLAlgorithm` with a huge amount of logging codes.

    The callback is able to produce the following outputs:

    - **Mask Figures**: both training and test, masks and cumulative masks.


    """

    def __init__(
        self,
        test_masks_plot_dir: str | None,
        test_cumulative_masks_plot_dir: str | None,
        training_masks_plot_dir: str | None,
        plot_training_mask_every_n_steps: int | None,
    ) -> None:
        r"""Initialise the `HATMetricsCallback`.

        **Args:**
        - **test_masks_plot_dir** (`str` | `None`): the directory to save the test mask figures. If `None`, no file will be saved.
        - **test_cumulative_masks_plot_dir** (`str` | `None`): the directory to save the test cumulative mask figures. If `None`, no file will be saved.
        - **training_masks_plot_dir** (`str` | `None`): the directory to save the training mask figures. If `None`, no file will be saved.
        - **plot_training_mask_every_n_steps** (`int`): the frequency of plotting training mask figures in terms of number of batches during training. Only applies when `training_masks_plot_dir` is not `None`.
        """
        Callback.__init__(self)

        # paths
        if not os.path.exists(test_masks_plot_dir):
            os.makedirs(test_masks_plot_dir, exist_ok=True)
        self.test_masks_plot_dir: str = test_masks_plot_dir
        r"""Store the directory to save the test mask figures."""
        if not os.path.exists(test_cumulative_masks_plot_dir):
            os.makedirs(test_cumulative_masks_plot_dir, exist_ok=True)
        self.test_cumulative_masks_plot_dir: str = test_cumulative_masks_plot_dir
        r"""Store the directory to save the test cumulative mask figures."""
        if not os.path.exists(training_masks_plot_dir):
            os.makedirs(training_masks_plot_dir, exist_ok=True)
        self.training_masks_plot_dir: str = training_masks_plot_dir
        r"""Store the directory to save the training mask figures."""

        # other settings
        self.plot_training_mask_every_n_steps: int = plot_training_mask_every_n_steps
        r"""Store the frequency of plotting training masks in terms of number of batches."""

        self.task_id: int
        r"""Task ID counter indicating which task is being processed. Self updated during the task loop."""

    def on_fit_start(self, trainer: Trainer, pl_module: CLAlgorithm) -> None:
        r"""Get the current task ID in the beginning of a task's fitting (training and validation). Sanity check the `pl_module` to be `HAT` or `AdaHAT`.

        **Raises:**
        -**TypeError**: when the `pl_module` is not `HAT` or `AdaHAT`.
        """

        # get the current task_id from the `CLAlgorithm` object
        self.task_id = pl_module.task_id

        # sanity check
        if not isinstance(pl_module, HAT):
            raise TypeError(
                "The `CLAlgorithm` should be `HAT` or `AdaHAT` to apply `HATMetricsCallback`!"
            )

    def on_train_batch_end(
        self,
        trainer: Trainer,
        pl_module: CLAlgorithm,
        outputs: dict[str, Any],
        batch: Any,
        batch_idx: int,
    ) -> None:
        r"""Plot training mask and log network capacity after training batch.

        **Args:**
        - **outputs** (`dict[str, Any]`): the outputs of the training step, which is the returns of the `training_step()` method in the `CLAlgorithm`.
        - **batch** (`Any`): the training data batch.
        - **batch_idx** (`int`): the index of the current batch. This is for the file name of mask figures.
        """

        # get the mask over the model after training the batch
        mask = outputs["mask"]
        # get the current network capacity
        capacity = outputs["capacity"]

        # plot the mask
        if batch_idx % self.plot_training_mask_every_n_steps == 0:
            plot.plot_hat_mask(
                mask=mask,
                plot_dir=self.training_masks_plot_dir,
                task_id=self.task_id,
                step=batch_idx,
            )

        # log the network capacity to Lightning loggers
        pl_module.log(
            f"task_{self.task_id}/train/network_capacity", capacity, prog_bar=True
        )

    def on_test_start(self, trainer: Trainer, pl_module: CLAlgorithm) -> None:
        r"""Plot test mask and cumulative mask figures."""

        # test mask
        mask = pl_module.masks[f"{self.task_id}"]
        plot.plot_hat_mask(
            mask=mask, plot_dir=self.test_masks_plot_dir, task_id=self.task_id
        )

        # cumulative mask
        cumulative_mask = pl_module.cumulative_mask_for_previous_tasks
        plot.plot_hat_mask(
            mask=cumulative_mask,
            plot_dir=self.test_cumulative_masks_plot_dir,
            task_id=self.task_id,
        )



r"""
The submodule in `callbacks` for `CLMetricsCallback`.
"""

__all__ = ["CLMetricsCallback"]

import logging
import os
from typing import Any

from lightning import Callback, Trainer

from clarena.cl_algorithms import CLAlgorithm
from clarena.utils import MeanMetricBatch, plot, save

# always get logger for built-in logging in each module
pylogger = logging.getLogger(__name__)


class CLMetricsCallback(Callback):
    r"""Provides all actions that are related to CL metrics, which include:

    - Defining, initialising and recording metrics.
    - Logging training and validation metrics to Lightning loggers in real time.
    - Saving test metrics to files.
    - Visualising test metrics as plots.

    Please refer to the [A Summary of Continual Learning Metrics](https://pengxiang-wang.com/posts/continual-learning-metrics) to learn what continual learning metrics mean.

    Lightning provides `self.log()` to log metrics in `LightningModule` where our `CLAlgorithm` based. You can put `self.log()` here if you don't want to mess up the `CLAlgorithm` with a huge amount of logging codes.

    The callback is able to produce the following outputs:

    - CSV files for test accuracy and classification loss (lower triangular) matrix, average accuracy and classification loss. See [here](https://pengxiang-wang.com/posts/continual-learning-metrics.html#sec-test-performance-of-previous-tasks) for details.
    - Coloured plot for test accuracy and classification loss (lower triangular) matrix. See [here](https://pengxiang-wang.com/posts/continual-learning-metrics.html#sec-test-performance-of-previous-tasks) for details.
    - Curve plots for test average accuracy and classification loss over different training tasks. See [here](https://pengxiang-wang.com/posts/continual-learning-metrics.html#sec-average-test-performance-over-tasks) for details.


    """

    def __init__(
        self,
        save_dir: str,
        test_acc_csv_name: str | None = None,
        test_loss_cls_csv_name: str | None = None,
        test_acc_matrix_plot_name: str | None = None,
        test_loss_cls_matrix_plot_name: str | None = None,
        test_ave_acc_plot_name: str | None = None,
        test_ave_loss_cls_plot_name: str | None = None,
    ) -> None:
        r"""Initialise the `CLMetricsCallback`.

        **Args:**
        - **save_dir** (`str` | `None`): the directory to save the test metrics files and plots. Better inside the output folder.
        - **test_acc_csv_name** (`str` | `None`): file name to save test accuracy matrix and average accuracy as CSV file. If `None`, no file will be saved.
        - **test_loss_cls_csv_name**(`str` | `None`): file name to save classification loss matrix and average classification loss as CSV file. If `None`, no file will be saved.
        - **test_acc_matrix_plot_name** (`str` | `None`): file name to save accuracy matrix plot. If `None`, no file will be saved.
        - **test_loss_cls_matrix_plot_name** (`str` | `None`): file name to save classification loss matrix plot. If `None`, no file will be saved.
        - **test_ave_acc_plot_name** (`str` | `None`): file name to save average accuracy as curve plot over different training tasks. If `None`, no file will be saved.
        - **test_ave_loss_cls_plot_name** (`str` | `None`): file name to save average classification loss as curve plot over different training tasks. If `None`, no file will be saved.
        """
        Callback.__init__(self)

        os.makedirs(save_dir, exist_ok=True)

        # paths
        self.test_acc_csv_path: str = os.path.join(save_dir, test_acc_csv_name)
        r"""Store the path to save test accuracy matrix and average accuracy CSV file."""
        self.test_loss_cls_csv_path: str = os.path.join(
            save_dir, test_loss_cls_csv_name
        )
        r"""Store the path to save test classification loss and average accuracy CSV file."""
        self.test_acc_matrix_plot_path: str = os.path.join(
            save_dir, test_acc_matrix_plot_name
        )
        r"""Store the path to save test accuracy matrix plot."""
        self.test_loss_cls_matrix_plot_path: str = os.path.join(
            save_dir, test_loss_cls_matrix_plot_name
        )
        r"""Store the path to save test classification loss matrix plot."""
        self.test_ave_acc_plot_path: str = os.path.join(
            save_dir, test_ave_acc_plot_name
        )
        r"""Store the path to save test average accuracy curve plot."""
        self.test_ave_loss_cls_plot_path: str = os.path.join(
            save_dir, test_ave_loss_cls_plot_name
        )
        r"""Store the path to save test average classification loss curve plot."""

        # training accumulated metrics
        self.acc_training_epoch: MeanMetricBatch
        r"""Classification accuracy of training epoch. Accumulated and calculated from the training batches. See [here](https://pengxiang-wang.com/posts/continual-learning-metrics.html#sec-performance-of-training-epoch) for details. """
        self.loss_cls_training_epoch: MeanMetricBatch
        r"""Classification loss of training epoch. Accumulated and calculated from the training batches. See [here](https://pengxiang-wang.com/posts/continual-learning-metrics.html#sec-performance-of-training-epoch) for details. """
        self.loss_training_epoch: MeanMetricBatch
        r"""Total loss of training epoch. Accumulated and calculated from the training batches. See [here](https://pengxiang-wang.com/posts/continual-learning-metrics.html#sec-performance-of-training-epoch) for details. """

        # validation accumulated metrics
        self.acc_val: MeanMetricBatch
        r"""Validation classification accuracy of the model after training epoch. Accumulated and calculated from the validation batches. See [here](https://pengxiang-wang.com/posts/continual-learning-metrics.html#sec-validation-performace) for details. """
        self.loss_cls_val: MeanMetricBatch
        r"""Validation classification of the model loss after training epoch. Accumulated and calculated from the validation batches. See [here](https://pengxiang-wang.com/posts/continual-learning-metrics.html#sec-validation-performace) for details. """

        # test accumulated metrics
        self.acc_test: dict[str, MeanMetricBatch]
        r"""Test classification accuracy of the current model (`self.task_id`) on current and previous tasks. Accumulated and calculated from the test batches. Keys are task IDs (string type) and values are the corresponding metrics. It is the last row of the lower triangular matrix. See [here](https://pengxiang-wang.com/posts/continual-learning-metrics.html#sec-test-performance-of-previous-tasks) for details. """
        self.loss_cls_test: dict[str, MeanMetricBatch]
        r"""Test classification loss of the current model (`self.task_id`) on current and previous tasks. Accumulated and calculated from the test batches. Keys are task IDs (string type) and values are the corresponding metrics. It is the last row of the lower triangular matrix. See [here](https://pengxiang-wang.com/posts/continual-learning-metrics.html#sec-test-performance-of-previous-tasks) for details. """

        self.task_id: int
        r"""Task ID counter indicating which task is being processed. Self updated during the task loop."""

    def on_fit_start(self, trainer: Trainer, pl_module: CLAlgorithm) -> None:
        r"""Initialise training and validation metrics."""

        # set the current task_id from the `CLAlgorithm` object
        self.task_id = pl_module.task_id

        # initialise training metrics
        self.loss_cls_training_epoch = MeanMetricBatch()
        self.loss_training_epoch = MeanMetricBatch()
        self.acc_training_epoch = MeanMetricBatch()

        # initialise validation metrics
        self.loss_cls_val = MeanMetricBatch()
        self.acc_val = MeanMetricBatch()

    def on_train_batch_end(
        self,
        trainer: Trainer,
        pl_module: CLAlgorithm,
        outputs: dict[str, Any],
        batch: Any,
        batch_idx: int,
    ) -> None:
        r"""Record training metrics from training batch, log metrics of training batch and accumulated metrics of the epoch to Lightning loggers.

        **Args:**
        - **outputs** (`dict[str, Any]`): the outputs of the training step, the returns of the `training_step()` method in the `CLAlgorithm`.
        - **batch** (`Any`): the training data batch.
        """
        # get the batch size
        batch_size = len(batch)

        # get training metrics values of current training batch from the outputs of the `training_step()`
        loss_cls_batch = outputs["loss_cls"]
        loss_batch = outputs["loss"]
        acc_batch = outputs["acc"]

        # update accumulated training metrics to calculate training metrics of the epoch
        self.loss_cls_training_epoch.update(loss_cls_batch, batch_size)
        self.loss_training_epoch.update(loss_cls_batch, batch_size)
        self.acc_training_epoch.update(acc_batch, batch_size)

        # log training metrics of current training batch to Lightning loggers
        pl_module.log(
            f"task_{self.task_id}/train/loss_cls_batch", loss_cls_batch, prog_bar=True
        )
        pl_module.log(
            f"task_{self.task_id}/train/loss_batch", loss_batch, prog_bar=True
        )
        pl_module.log(f"task_{self.task_id}/train/acc_batch", acc_batch, prog_bar=True)

        # log accumulated training metrics till this training batch to Lightning loggers
        pl_module.log(
            f"task_{self.task_id}/train/loss_cls",
            self.loss_cls_training_epoch.compute(),
            prog_bar=True,
        )
        pl_module.log(
            f"task_{self.task_id}/train/loss",
            self.loss_training_epoch.compute(),
            prog_bar=True,
        )
        pl_module.log(
            f"task_{self.task_id}/train/acc",
            self.acc_training_epoch.compute(),
            prog_bar=True,
        )

    def on_train_epoch_end(
        self,
        trainer: Trainer,
        pl_module: CLAlgorithm,
    ) -> None:
        r"""Log metrics of training epoch to plot learning curves and reset the metrics accumulation at the end of training epoch."""

        # log the accumulated and computed metrics of the epoch to Lightning loggers, specially for plotting learning curves
        pl_module.log(
            f"task_{self.task_id}/learning_curve/train/loss_cls",
            self.loss_cls_training_epoch.compute(),
            on_epoch=True,
            prog_bar=True,
        )
        pl_module.log(
            f"task_{self.task_id}/learning_curve/train/acc",
            self.acc_training_epoch.compute(),
            on_epoch=True,
            prog_bar=True,
        )

        # reset the metrics of training epoch as there are more epochs to go and not only one epoch like in the validation and test
        self.loss_cls_training_epoch.reset()
        self.loss_training_epoch.reset()
        self.acc_training_epoch.reset()

    def on_validation_batch_end(
        self,
        trainer: Trainer,
        pl_module: CLAlgorithm,
        outputs: dict[str, Any],
        batch: Any,
        batch_idx: int,
    ) -> None:
        r"""Accumulating metrics from validation batch. We don't need to log and monitor the metrics of validation batches.

        **Args:**
        - **outputs** (`dict[str, Any]`): the outputs of the validation step, which is the returns of the `validation_step()` method in the `CLAlgorithm`.
        - **batch** (`Any`): the validation data batch.
        """

        # get the batch size
        batch_size = len(batch)

        # get the metrics values of the batch from the outputs
        loss_cls_batch = outputs["loss_cls"]
        acc_batch = outputs["acc"]

        # update the accumulated metrics in order to calculate the validation metrics
        self.loss_cls_val.update(loss_cls_batch, batch_size)
        self.acc_val.update(acc_batch, batch_size)

    def on_validation_epoch_end(
        self,
        trainer: Trainer,
        pl_module: CLAlgorithm,
    ) -> None:
        r"""Log validation metrics to plot learning curves."""

        # log the accumulated and computed metrics of the epoch to Lightning loggers, specially for plotting learning curves
        pl_module.log(
            f"task_{self.task_id}/learning_curve/val/loss_cls",
            self.loss_cls_val.compute(),
            on_epoch=True,
            prog_bar=True,
        )
        pl_module.log(
            f"task_{self.task_id}/learning_curve/val/acc",
            self.acc_val.compute(),
            on_epoch=True,
            prog_bar=True,
        )

    def on_test_start(
        self,
        trainer: Trainer,
        pl_module: CLAlgorithm,
    ) -> None:
        r"""Initialise the metrics for testing each seen task in the beginning of a task's testing."""

        # set the current task_id again (double checking) from the `CLAlgorithm` object
        self.task_id = pl_module.task_id

        # initialise test metrics for current and previous tasks
        self.loss_cls_test = {
            f"{task_id}": MeanMetricBatch() for task_id in range(1, self.task_id + 1)
        }
        self.acc_test = {
            f"{task_id}": MeanMetricBatch() for task_id in range(1, self.task_id + 1)
        }

    def on_test_batch_end(
        self,
        trainer: Trainer,
        pl_module: CLAlgorithm,
        outputs: dict[str, Any],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        r"""Accumulating metrics from test batch. We don't need to log and monitor the metrics of test batches.

        **Args:**
        - **outputs** (`dict[str, Any]`): the outputs of the test step, which is the returns of the `test_step()` method in the `CLAlgorithm`.
        - **batch** (`Any`): the validation data batch.
        - **dataloader_idx** (`int`): the task ID of seen tasks to be tested. A default value of 0 is given otherwise the LightningModule will raise a `RuntimeError`.
        """

        # get the batch size
        batch_size = len(batch)

        task_id = dataloader_idx + 1  # our `task_id` system starts from 1

        # get the metrics values of the batch from the outputs
        loss_cls_batch = outputs["loss_cls"]
        acc_batch = outputs["acc"]

        # update the accumulated metrics in order to calculate the metrics of the epoch
        self.loss_cls_test[f"{task_id}"].update(loss_cls_batch, batch_size)
        self.acc_test[f"{task_id}"].update(acc_batch, batch_size)

    def on_test_epoch_end(
        self,
        trainer: Trainer,
        pl_module: CLAlgorithm,
    ) -> None:
        r"""Save and plot test metrics at the end of test."""

        # save (update) the test metrics to CSV files
        save.update_test_acc_to_csv(
            test_acc_metric=self.acc_test,
            csv_path=self.test_acc_csv_path,
        )
        save.update_test_loss_cls_to_csv(
            test_loss_cls_metric=self.loss_cls_test,
            csv_path=self.test_loss_cls_csv_path,
        )

        # plot the test metrics
        plot.plot_test_acc_matrix_from_csv(
            csv_path=self.test_acc_csv_path,
            task_id=self.task_id,
            plot_path=self.test_acc_matrix_plot_path,
        )
        plot.plot_test_loss_cls_matrix_from_csv(
            csv_path=self.test_loss_cls_csv_path,
            task_id=self.task_id,
            plot_path=self.test_loss_cls_matrix_plot_path,
        )
        plot.plot_test_ave_acc_curve_from_csv(
            csv_path=self.test_acc_csv_path,
            task_id=self.task_id,
            plot_path=self.test_ave_acc_plot_path,
        )
        plot.plot_test_ave_loss_cls_curve_from_csv(
            csv_path=self.test_loss_cls_csv_path,
            task_id=self.task_id,
            plot_path=self.test_ave_loss_cls_plot_path,
        )



r"""
The submodule in `cl_algorithms` for [CBP (Continual Backpropagation)](https://www.nature.com/articles/s41586-024-07711-7) algorithm.
"""

__all__ = ["CBP"]

import logging
from typing import Any

import torch
from torch import Tensor

from clarena.backbones import CLBackbone
from clarena.cl_algorithms import Finetuning
from clarena.cl_heads import HeadsCIL, HeadsTIL
from clarena.utils.transforms import min_max_normalise

# always get logger for built-in logging in each module
pylogger = logging.getLogger(__name__)


class CBP(Finetuning):
    r"""CBP (Continual Backpropagation) algorithm.

    [CBP (Continual Backpropagation, 2024)](https://www.nature.com/articles/s41586-024-07711-7) is a continual learning approach that reinitialises a small number of units during training, using an utility measures to determine which units to reinitialise. It aims to address loss of plasticity problem for learning new tasks, yet not very well solve the catastrophic forgetting problem in continual learning.

    We implement CBP as a subclass of Finetuning algorithm, as CBP has the same `forward()`, `training_step()`, `validation_step()` and `test_step()` method as `Finetuning` class.
    """

    def __init__(
        self,
        backbone: CLBackbone,
        heads: HeadsTIL | HeadsCIL,
        replacement_rate: float,
        maturity_threshold: int,
        utility_decay_rate: float,
    ) -> None:
        r"""Initialise the Finetuning algorithm with the network. It has no additional hyperparamaters.

        **Args:**
        - **backbone** (`CLBackbone`): backbone network.
        - **heads** (`HeadsTIL` | `HeadsCIL`): output heads.
        - **replacement_rate** (`float`): the replacement rate of units. It is the precentage of units to be reinitialised during training.
        - **maturity_threshold** (`int`): the maturity threshold of units. It is the number of training steps before a unit can be reinitialised.
        - **utility_decay_rate** (`float`): the utility decay rate of units. It is the rate at which the utility of a unit decays over time.
        """
        Finetuning.__init__(self, backbone=backbone, heads=heads)

        self.replacement_rate: float = replacement_rate
        r"""Store the replacement rate of units. """
        self.maturity_threshold: int = maturity_threshold
        r"""Store the maturity threshold of units. """
        self.utility_decay_rate: float = utility_decay_rate
        r"""Store the utility decay rate of units. """

        self.contribution_utility: dict[str, Tensor] = {}
        r"""Store the contribution utility of units. See equation (1) in the [continual backpropagation paper](https://www.nature.com/articles/s41586-024-07711-7). Keys are layer names and values are the utility tensor for the layer. The utility tensor is the same size as the feature tensor with size (number of units). """
        self.num_replacements: dict[str, int] = {}
        r"""Store the number of replacements of units in each layer. Keys are layer names and values are the number of replacements for the layer. """
        self.age: dict[str, Tensor] = {}
        r"""Store the age of units. Keys are layer names and values are the age tensor for the layer. The age tensor is the same size as the feature tensor with size (1, number of units). """

    def on_train_start(self) -> None:
        r"""Initialise the utility, number of replacements and age for each layer as zeros."""

        # initialise the utility, number of replacements and age as zeros at the beginning of first task. This should not be called in `__init__()` method as the `self.device` is not available at that time.
        if self.task_id == 1:
            for layer_name in self.backbone.weighted_layer_names:
                layer = self.backbone.get_layer_by_name(
                    layer_name
                )  # get the layer by its name
                num_units = layer.weight.shape[0]

                self.contribution_utility[layer_name] = torch.zeros(num_units).to(
                    self.device
                )
                self.num_replacements[layer_name] = 0
                self.age[layer_name] = torch.zeros(num_units).to(self.device)

    def on_train_batch_end(
        self, outputs: dict[str, Any], batch: Any, batch_idx: int
    ) -> None:
        r"""Update the contribution utility and age of units after each training step, and conduct reinitialisation of units based on utility measures. This is the core of the CBP algorithm.

        **Args:**
        - **outputs** (`dict[str, Any]`): the outputs of the training step, which is the returns of the `training_step()` method in the `CLAlgorithm`.
        - **batch** (`Any`): the training data batch.
        - **batch_idx** (`int`): the index of the current batch. This is for the file name of mask figures.
        """

        hidden_features = outputs["hidden_features"]

        for layer_name in self.backbone.weighted_layer_names:
            # layer-wise operation

            layer = self.backbone.get_layer_by_name(
                layer_name
            )  # get the layer by its name

            # update age
            self.age[layer_name] += 1

            # calculate current contribution utility
            current_contribution_utility = (
                torch.mean(
                    torch.abs(hidden_features[layer_name]),
                    dim=0,  # average the features over batch samples
                )
                * torch.sum(
                    torch.abs(layer.weight),
                    dim=1,  # sum over the output dimension
                )
            ).detach()
            current_contribution_utility = min_max_normalise(
                current_contribution_utility
            )  # normalise the utility to [0,1] to avoid linearly increasing utility

            # update utility
            self.contribution_utility[layer_name] = (
                self.utility_decay_rate * self.contribution_utility[layer_name]
                + (1 - self.utility_decay_rate) * current_contribution_utility
            )

            # find eligible units
            eligible_mask = self.age[layer_name] > self.maturity_threshold
            eligible_indices = torch.where(eligible_mask)[0]

            # update the number of replacements
            num_eligible_units = eligible_indices.numel()
            self.num_replacements[layer_name] += int(
                self.replacement_rate * num_eligible_units
            )

            # if the number of replacements is greater than 1, execute the replacement
            if self.num_replacements[layer_name] > 1:

                # find the unit with smallest utility among eligible units
                replaced_unit_idx = eligible_indices[
                    torch.argmin(
                        self.contribution_utility[layer_name][eligible_indices]
                        / self.age[layer_name][eligible_indices]
                    ).item()
                ]

                # reinitialise the input weights of the unit
                preceding_layer_name = self.backbone.preceding_layer_name(layer_name)
                if preceding_layer_name is not None:
                    preceding_layer = self.backbone.get_layer_by_name(
                        preceding_layer_name
                    )
                    with torch.no_grad():

                        preceding_layer.weight[:, replaced_unit_idx] = torch.rand_like(
                            preceding_layer.weight[:, replaced_unit_idx]
                        )

                # reinitalise the output weights of the unit
                with torch.no_grad():
                    layer.weight[replaced_unit_idx] = torch.rand_like(
                        layer.weight[replaced_unit_idx]
                    )

                # reinitialise utility
                self.contribution_utility[layer_name][replaced_unit_idx] = 0.0

                # reintialise age
                self.age[layer_name][replaced_unit_idx] = 0

                # update the number of replacements
                self.num_replacements[layer_name] -= 1



r"""
The submodule in `callbacks` for `PyloggerCallback`.
"""

__all__ = ["PyloggerCallback"]

import logging

from lightning import Callback, Trainer

from clarena.cl_algorithms import CLAlgorithm

# always get logger for built-in logging in each module
pylogger = logging.getLogger(__name__)


class PyloggerCallback(Callback):
    r"""Pylogger Callback provides additional logging for during continual learning progress.

    Put logging messages here if you don't want to mess up the `CLAlgorithm` (`LightningModule`) with a huge amount of logging codes.
    """

    def on_fit_start(self, trainer: Trainer, pl_module: CLAlgorithm) -> None:
        r"""Log messages for the start of training task."""
        pylogger.info("Start training task %s!", pl_module.task_id)

    def on_train_end(self, trainer: Trainer, pl_module: CLAlgorithm) -> None:
        r"""Log messages for the end of training task."""
        pylogger.info("Finish training task %s!", pl_module.task_id)

    def on_test_start(self, trainer: Trainer, pl_module: CLAlgorithm) -> None:
        r"""Log messages for the start of testing task."""
        pylogger.info(
            "Start testing task %s on all previous and current tasks!",
            pl_module.task_id,
        )

    def on_test_end(self, trainer: Trainer, pl_module: CLAlgorithm) -> None:
        r"""Log messages for the end of testing task."""
        pylogger.info(
            "Finish testing task %s on all previous and current tasks!",
            pl_module.task_id,
        )



r"""
The submodule in `backbones` for ResNet backbone network.
"""

__all__ = [
    "ResNetBlockSmall",
    "ResNetBlockLarge",
    "ResNetBase",
    "ResNet18",
    "ResNet34",
    "ResNet50",
    "ResNet101",
    "ResNet152",
    "HATMaskResNetBlockSmall",
    "HATMaskResNetBlockLarge",
    "HATMaskResNetBase",
    "HATMaskResNet18",
    "HATMaskResNet34",
    "HATMaskResNet50",
    "HATMaskResNet101",
    "HATMaskResNet152",
]


from torch import Tensor, nn

from clarena.backbones import CLBackbone, HATMaskBackbone


class ResNetBlockSmall(CLBackbone):
    r"""The smaller building block for ResNet-18/34.

    It consists of 2 weight convolutional layers, each followed by an activation function. See Table 1 or Figure 5 (left) in the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html).
    """

    def __init__(
        self,
        outer_layer_name: str,
        block_idx: int,
        preceding_output_channels: int,
        input_channels: int,
        overall_stride: int,
        activation_layer: nn.Module | None = nn.ReLU,
        batch_normalisation: bool = True,
        bias: bool = False,
    ) -> None:
        r"""Construct and initialise the smaller building block.

        **Args:**
        - **outer_layer_name** (`str`): pass the name of the multi-building-block layer that contains this building block to construct the full name of each weighted convolutional layer.
        - **block_idx** (`int`): the index of the building blocks in the multi-building-block layer to construct the full name of each weighted convolutional layer.
        - **preceding_output_channels** (`int`): the number of channels of preceding output of this particular building block.
        - **input_channels** (`int`): the number of channels of input of this building block.
        - **overall_stride** (`int`): the overall stride of this building block. This stride is performed at 2nd (last) convolutional layer where the 1st convolutional layer remain stride of 1.
        - **activation_layer** (`nn.Module`): activation function of each layer (if not `None`), if `None` this layer won't be used. Default `nn.ReLU`.
        - **batch_normalisation** (`bool`): whether to use batch normalisation after the weight convolutional layers. Default `True`, same as what the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html) does.
        - **bias** (`bool`): whether to use bias in the convolutional layer. Default `False`, because batch normalisation are doing the similar thing with bias.
        """
        CLBackbone.__init__(self, output_dim=None)

        self.batch_normalisation: bool = batch_normalisation
        r"""Store whether to use batch normalisation after the fully-connected layers."""
        self.activation: bool = activation_layer is not None
        r"""Store whether to use activation function after the fully-connected layers."""

        self.full_1st_layer_name = f"{outer_layer_name}/{block_idx}/conv1"
        r"""Format and store full name of the 1st weighted convolutional layer. """
        self.full_2nd_layer_name = f"{outer_layer_name}/{block_idx}/conv2"
        r"""Format and store full name of the 2nd weighted convolutional layer. """

        # construct the 1st weighted convolutional layer and attached layers (batchnorm, activation, etc)
        layer_input_channels = preceding_output_channels  # the input channels of the 1st convolutional layer, which receive the output channels of the preceding module
        layer_output_channels = (
            input_channels  # the output channels of the 1st convolutional layer
        )
        self.conv1 = nn.Conv2d(
            in_channels=layer_input_channels,
            out_channels=layer_output_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=bias,
        )  # construct the 1st weight convolutional layer of the smaller building block. Overall stride is not performed here
        r"""The 1st weight convolutional layer of the smaller building block. """
        self.weighted_layer_names.append(
            self.full_1st_layer_name
        )  # update the weighted layer names
        if self.batch_normalisation:
            self.conv_bn1 = nn.BatchNorm2d(
                num_features=layer_output_channels
            )  # construct the batch normalisation layer
            r"""The batch normalisation (`nn.BatchNorm2d`) layer after the 1st weighted convolutional layer. """
        if self.activation:
            self.conv_activation1 = activation_layer()  # construct the activation layer
            r"""The activation layer after the 1st weighted convolutional layer. """

        # construct the 2nd weighted convolutional layer and attached layers (batchnorm, activation, etc)
        layer_input_channels = input_channels  # the input channels of the 2nd convolutional layer, which is `input_channels`, the same as the output channels of the 1st convolutional layer
        layer_output_channels = (
            input_channels * 1
        )  # the output channels of the 2nd convolutional layer, which is the same as the input channels (without expansion)
        self.conv2 = nn.Conv2d(
            in_channels=layer_input_channels,
            out_channels=layer_output_channels,
            kernel_size=3,
            stride=overall_stride,
            padding=1,
            bias=bias,
        )  # construct the 2nd weight convolutional layer of the smaller building block. Overall stride is performed here
        r"""The 2nd weight convolutional layer of the smaller building block. """
        self.weighted_layer_names.append(
            self.full_2nd_layer_name
        )  # update the weighted layer names
        if batch_normalisation:
            self.conv_bn2 = nn.BatchNorm2d(
                num_features=layer_output_channels
            )  # construct the batch normalisation layer
            r"""The batch normalisation (`nn.BatchNorm2d`) layer after the 2nd weighted convolutional layer. """
        if self.activation:
            self.conv_activation2 = activation_layer()  # construct the activation layer
            r"""The activation layer after the 2nd weighted convolutional layer. """

        self.identity_downsample: nn.Module = (
            nn.Conv2d(
                in_channels=preceding_output_channels,
                out_channels=input_channels,
                kernel_size=1,
                stride=overall_stride,
                bias=False,
            )
            if preceding_output_channels != input_channels or overall_stride != 1
            else None
        )  # construct the identity downsample function
        r"""The convolutional layer for downsampling identity in the shortcut connection if the dimension of identity from input doesn't match the output's. This case only happens when the number of input channels doesn't equal to the number of preceding output channels or a layer with stride > 1 exists. """

    def forward(self, input: Tensor) -> tuple[Tensor, dict[str, Tensor]]:
        r"""The forward pass for data.

        **Args:**
        - **input** (`Tensor`): the input feature maps.

        **Returns:**
        - **output_feature** (`Tensor`): the output feature maps.
        - **hidden_features** (`dict[str, Tensor]`): the hidden features (after activation) in each weighted layer. Key (`str`) is the weighted layer name, value (`Tensor`) is the hidden feature tensor. This is used for the continual learning algorithms that need to use the hidden features for various purposes.
        """
        hidden_features = {}

        identity = (
            self.identity_downsample(input)
            if self.identity_downsample is not None
            else input
        )  # remember the identity of input for the shortcut connection. Perform downsampling if its dimension doesn't match the output's

        x = input
        x = self.conv1(x)
        if self.batch_normalisation:
            x = self.conv_bn1(x)
        if self.activation:
            x = self.conv_activation1(x)
        hidden_features[self.full_1st_layer_name] = x  # store the hidden feature

        x = self.conv2(x)
        if self.batch_normalisation:
            x = self.conv_bn2(x)

        x = x + identity
        if self.activation:
            x = self.conv_activation2(x)  # activation after the shortcut connection
        hidden_features[self.full_2nd_layer_name] = x  # store the hidden feature

        output_feature = x

        return output_feature, hidden_features


class ResNetBlockLarge(CLBackbone):
    r"""The larger building block for ResNet-50/101/152. It is referred to "bottleneck" building block in the paper.

    It consists of 3 weight convolutional layers, each followed by an activation function. See Table 1 or Figure 5 (right) in the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html).
    """

    def __init__(
        self,
        outer_layer_name: str,
        block_idx: int,
        preceding_output_channels: int,
        input_channels: int,
        overall_stride: int,
        activation_layer: nn.Module | None = nn.ReLU,
        batch_normalisation: bool = True,
        bias: bool = False,
    ) -> None:
        r"""Construct and initialise the larger building block.

        **Args:**
        - **outer_layer_name** (`str`): pass the name of the multi-building-block layer that contains this building block to construct the full name of each weighted convolutional layer.
        - **block_idx** (`int`): the index of the building blocks in the multi-building-block layer to construct the full name of each weighted convolutional layer.
        - **preceding_output_channels** (`int`): the number of channels of preceding output of this particular building block.
        - **input_channels** (`int`): the number of channels of input of this building block.
        - **overall_stride** (`int`): the overall stride of this building block. This stride is performed at 2nd (middle) convolutional layer where 1st and 3rd convolutional layers remain stride of 1.
        - **activation_layer** (`nn.Module`): activation function of each layer (if not `None`), if `None` this layer won't be used. Default `nn.ReLU`.
        - **batch_normalisation** (`bool`): whether to use batch normalisation after the weight convolutional layers. Default `True`, same as what the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html) does.
        - **bias** (`bool`): whether to use bias in the convolutional layer. Default `False`, because batch normalisation are doing the similar thing with bias.
        """
        CLBackbone.__init__(self, output_dim=None)

        self.batch_normalisation: bool = batch_normalisation
        r"""Store whether to use batch normalisation after the fully-connected layers."""
        self.activation: bool = activation_layer is not None
        r"""Store whether to use activation function after the fully-connected layers."""

        self.full_1st_layer_name = f"{outer_layer_name}/{block_idx}/conv1"
        r"""Format and store full name of the 1st weighted convolutional layer. """
        self.full_2nd_layer_name = f"{outer_layer_name}_{block_idx}_conv2"
        r"""Format and store full name of the 2nd weighted convolutional layer. """
        self.full_3rd_layer_name = f"{outer_layer_name}_{block_idx}_conv3"
        r"""Format and store full name of the 3rd weighted convolutional layer. """

        # construct the 1st weighted convolutional layer and attached layers (batchnorm, activation, etc)
        layer_input_channels = preceding_output_channels  # the input channels of the 1st convolutional layer, which receive the output channels of the preceding module
        layer_output_channels = (
            input_channels  # the output channels of the 1st convolutional layer
        )
        self.conv1 = nn.Conv2d(
            in_channels=layer_input_channels,
            out_channels=layer_output_channels,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=bias,
        )  # construct the 1st weight convolutional layer of the larger building block. Overall stride is not performed here
        r"""The 1st weight convolutional layer of the larger building block. """
        self.weighted_layer_names.append(
            self.full_1st_layer_name
        )  # update the weighted layer names
        if self.batch_normalisation:
            self.conv_bn1 = nn.BatchNorm2d(
                num_features=layer_output_channels
            )  # construct the batch normalisation layer
            r"""The batch normalisation (`nn.BatchNorm2d`) layer after the 1st weighted convolutional layer. """
        if self.activation:
            self.conv_activation1 = activation_layer()  # construct the activation layer
            r"""The activation layer after the 1st weighted convolutional layer. """

        # construct the 2nd weighted convolutional layer and attached layers (batchnorm, activation, etc)
        layer_input_channels = input_channels  # the input channels of the 2nd convolutional layer, which is `input_channels`, the same as the output channels of the 1st convolutional layer
        layer_output_channels = (
            input_channels
            * 1  # the output channels of the 2nd convolutional layer, which is the same as the input channels (without expansion)
        )
        self.conv2 = nn.Conv2d(
            in_channels=layer_input_channels,
            out_channels=layer_output_channels,
            kernel_size=3,
            stride=overall_stride,
            padding=1,
            bias=bias,
        )  # construct the 2nd weight convolutional layer of the larger building block. Overall stride is performed here
        r"""The 2nd weight convolutional layer of the larger building block. """
        self.weighted_layer_names.append(
            self.full_2nd_layer_name
        )  # update the weighted layer names
        if self.batch_normalisation:
            self.conv_bn2 = nn.BatchNorm2d(
                num_features=layer_output_channels
            )  # construct the batch normalisation layer
            r"""The batch normalisation (`nn.BatchNorm2d`) layer after the 2nd weighted convolutional layer. """
        if self.activation:
            self.conv_activation2 = activation_layer()  # construct the activation layer
            r"""The activation layer after the 2nd weighted convolutional layer. """

        # construct the 3rd weighted convolutional layer and attached layers (batchnorm, activation, etc)
        layer_input_channels = (
            input_channels * 1
        )  # the input channels of the 2nd convolutional layer, which is `input_channels * 1`, the same as the output channels of the 1st convolutional layer
        layer_output_channels = (
            input_channels
            * 4  # the output channels of the 2nd convolutional layer, which is 4 times expanded as the input channels
        )
        self.conv3 = nn.Conv2d(
            in_channels=layer_input_channels,
            out_channels=layer_output_channels,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=bias,
        )  # construct the 3rd weight convolutional layer of the larger building block. Overall stride is not performed here
        r"""The 3rd weight convolutional layer of the larger building block. """
        self.weighted_layer_names.append(
            self.full_3rd_layer_name
        )  # update the weighted layer names
        if batch_normalisation:
            self.conv_bn3 = nn.BatchNorm2d(
                num_features=layer_output_channels
            )  # construct the batch normalisation layer
            r"""The batch normalisation (`nn.BatchNorm2d`) layer after the 3rd weighted convolutional layer. """
        if self.activation:
            self.conv_activation3 = activation_layer()  # construct the activation layer
            r"""The activation layer after the 3rd weighted convolutional layer. """

        self.identity_downsample: nn.Module = (
            nn.Conv2d(
                in_channels=preceding_output_channels,
                out_channels=input_channels * 4,
                kernel_size=1,
                stride=overall_stride,
                bias=False,
            )
            if preceding_output_channels != input_channels * 4 or overall_stride != 1
            else None
        )
        r"""The convolutional layer for downsampling identity in the shortcut connection if the dimension of identity from input doesn't match the output's. This case only happens when the number of input channels doesn't equal to the number of preceding output channels or a layer with stride > 1 exists. """

    def forward(self, input: Tensor) -> tuple[Tensor, dict[str, Tensor]]:
        r"""The forward pass for data.

        **Args:**
        - **input** (`Tensor`): the input feature maps.

        **Returns:**
        - **output_feature** (`Tensor`): the output feature maps.
        - **hidden_features** (`dict[str, Tensor]`): the hidden features (after activation) in each weighted layer. Key (`str`) is the weighted layer name, value (`Tensor`) is the hidden feature tensor. This is used for the continual learning algorithms that need to use the hidden features for various purposes.
        """
        hidden_features = {}

        identity = (
            self.identity_downsample(input)
            if self.identity_downsample is not None
            else input
        )  # remember the identity of input for the shortcut connection. Perform downsampling if its dimension doesn't match the output's

        x = input
        x = self.conv1(x)
        if self.batch_normalisation:
            x = self.conv_bn1(x)
        if self.activation:
            x = self.conv_activation1(x)
        hidden_features[self.full_1st_layer_name] = x  # store the hidden feature

        x = self.conv2(x)
        if self.batch_normalisation:
            x = self.conv_bn2(x)
        if self.activation:
            x = self.conv_activation2(x)
        hidden_features[self.full_2nd_layer_name] = x  # store the hidden feature

        x = self.conv3(x)
        if self.batch_normalisation:
            x = self.conv_bn3(x)

        x = x + identity
        if self.activation:
            x = self.conv_activation3(x)  # activation after the shortcut connection
        hidden_features[self.full_3rd_layer_name] = x  # store the hidden feature

        output_feature = x

        return output_feature, hidden_features


class ResNetBase(CLBackbone):
    r"""The base class of [residual network (ResNet)](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html).

    ResNet is a convolutional network architecture, which has 1st convolutional parameter layer and a maxpooling layer, connecting to 4 convolutional layers which contains multiple convolutional parameter layer. Each layer of the 4 are constructed from basic building blocks which are either small (`ResNetBlockSmall`) or large (`ResNetBlockLarge`). Each building block contains several convolutional parameter layers. The building blocks are connected by a skip connection which is a direct connection from the input of the block to the output of the block, and this is why it's called residual (find "shortcut connections" in the paper for more details). After the 5th convolutional layer, there are average pooling layer and a fully connected layer which connects to the CL output heads.
    """

    def __init__(
        self,
        input_channels: int,
        building_block_type: ResNetBlockSmall | ResNetBlockLarge,
        building_block_nums: tuple[int, int, int, int],
        building_block_preceding_output_channels: tuple[int, int, int, int],
        building_block_input_channels: tuple[int, int, int, int],
        output_dim: int,
        activation_layer: nn.Module | None = nn.ReLU,
        batch_normalisation: bool = True,
        bias: bool = False,
    ) -> None:
        r"""Construct and initialise the ResNet backbone network.

        **Args:**
        - **input_channels** (`int`): the number of channels of input. Image data are kept channels when going in ResNet. Note that convolutional networks require number of input channels instead of dimension.
        - **building_block_type** (`ResNetBlockSmall` | `ResNetBlockLarge`): the type of building block used in the ResNet.
        - **building_block_nums** (`tuple[int, int, int, int]`): the number of building blocks in the 2-5 convolutional layer correspondingly.
        - **building_block_preceding_output_channels** (`tuple[int, int, int, int]`): the number of channels of preceding output of each building block in the 2-5 convolutional layer correspondingly.
        - **building_block_input_channels** (`tuple[int, int, int, int]`): the number of channels of input of each building block in the 2-5 convolutional layer correspondingly.
        - **output_dim** (`int`): the output dimension after flattening at last which connects to CL output heads. Although this is not determined by us but the architecture built before the flattening layer, we still need to provide this to construct the heads.
        - **activation_layer** (`nn.Module`): activation function of each layer (if not `None`), if `None` this layer won't be used. Default `nn.ReLU`.
        - **batch_normalisation** (`bool`): whether to use batch normalisation after the weight convolutional layers. Default `True`, same as what the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html) does.
        - **bias** (`bool`): whether to use bias in the convolutional layer. Default `False`, because batch normalisation are doing the similar thing with bias.
        """
        CLBackbone.__init__(self, output_dim=output_dim)

        self.batch_normalisation: bool = batch_normalisation
        r"""Store whether to use batch normalisation after the fully-connected layers."""
        self.activation: bool = activation_layer is not None
        r"""Store whether to use activation function after the fully-connected layers."""

        # construct the 1st weighted convolutional layer and attached layers (batchnorm, activation, etc)
        layer_input_channels = input_channels  # the input channels of the 1st convolutional layer, which receive the input of the entire network
        layer_output_channels = 64  # the output channels of the 1st convolutional layer
        self.conv1 = nn.Conv2d(
            in_channels=layer_input_channels,
            out_channels=layer_output_channels,
            kernel_size=7,
            stride=2,
            padding=3,
            bias=bias,
        )  # construct the 1st weight convolutional layer of the entire ResNet
        r"""The 1st weight convolutional layer of the entire ResNet. It  is always with fixed kernel size 7x7, stride 2, and padding 3. """
        self.weighted_layer_names.append("conv1")  # collect the layer name to be masked
        if self.batch_normalisation:
            self.conv_bn1 = nn.BatchNorm2d(
                num_features=layer_output_channels
            )  # construct the batch normalisation layer
            r"""The batch normalisation (`nn.BatchNorm2d`) layer after the 1st weighted convolutional layer. """
        if self.activation:
            self.conv_activation1 = activation_layer()  # construct the activation layer
            r"""The activation layer after the 1st weighted convolutional layer. """

        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)  #
        r"""The max pooling layer which is laid in between 1st and 2nd convolutional layers with kernel size 3x3, stride 2. """

        # construct the 2nd convolutional layer with multiple blocks, and its attached layers (batchnorm, activation, etc)
        self.conv2x = self._multiple_blocks(
            layer_name="conv2x",
            building_block_type=building_block_type,
            building_block_num=building_block_nums[0],
            preceding_output_channels=building_block_preceding_output_channels[0],
            input_channels=building_block_input_channels[0],
            overall_stride=1,  # the overall stride of the 2nd convolutional layer should be 1, as the preceding maxpooling layer has stride 2, which already made 112x112 -> 56x56. See Table 2 in the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html) for details.
            activation_layer=activation_layer,
            batch_normalisation=batch_normalisation,
            bias=bias,
        )
        r"""The 2nd convolutional layer of the ResNet, which contains multiple blocks. """

        # construct the 3rd convolutional layer with multiple blocks, and its attached layers (batchnorm, activation, etc)
        self.conv3x = self._multiple_blocks(
            layer_name="conv3x",
            building_block_type=building_block_type,
            building_block_num=building_block_nums[1],
            preceding_output_channels=building_block_preceding_output_channels[1],
            input_channels=building_block_input_channels[1],
            overall_stride=2,  # the overall stride of the 3rd convolutional layer should be 2, making 56x56 -> 28x28. See Table 2 in the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html) for details.
            activation_layer=activation_layer,
            batch_normalisation=batch_normalisation,
            bias=bias,
        )
        r"""The 3rd convolutional layer of the ResNet, which contains multiple blocks. """

        # construct the 4th convolutional layer with multiple blocks, and its attached layers (batchnorm, activation, etc)
        self.conv4x = self._multiple_blocks(
            layer_name="conv4x",
            building_block_type=building_block_type,
            building_block_num=building_block_nums[2],
            preceding_output_channels=building_block_preceding_output_channels[2],
            input_channels=building_block_input_channels[2],
            overall_stride=2,  # the overall stride of the 4th convolutional layer should be 2, making 28x28 -> 14x14. See Table 2 in the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html) for details.
            activation_layer=activation_layer,
            batch_normalisation=batch_normalisation,
            bias=bias,
        )
        r"""The 4th convolutional layer of the ResNet, which contains multiple blocks. """

        # construct the 5th convolutional layer with multiple blocks, and its attached layers (batchnorm, activation, etc)
        self.conv5x = self._multiple_blocks(
            layer_name="conv5x",
            building_block_type=building_block_type,
            building_block_num=building_block_nums[3],
            preceding_output_channels=building_block_preceding_output_channels[3],
            input_channels=building_block_input_channels[3],
            overall_stride=2,  # the overall stride of the 2nd convolutional layer should be 2, making 14x14 -> 7x7. See Table 2 in the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html) for details.
            activation_layer=activation_layer,
            batch_normalisation=batch_normalisation,
            bias=bias,
        )
        r"""The 5th convolutional layer of the ResNet, which contains multiple blocks. """

        self.avepool = nn.AdaptiveAvgPool2d(output_size=(1, 1))
        r"""The average pooling layer which is laid after the convolutional layers and before feature maps are flattened. """

    def _multiple_blocks(
        self,
        layer_name: str,
        building_block_type: ResNetBlockSmall | ResNetBlockLarge,
        building_block_num: int,
        preceding_output_channels: int,
        input_channels: int,
        overall_stride: int,
        activation_layer: nn.Module | None = nn.ReLU,
        batch_normalisation: bool = True,
        bias: bool = False,
    ) -> nn.Sequential:
        r"""Construct a layer consisting of multiple building blocks. It's used to construct the 2-5 convolutional layers of the ResNet.

        The "shortcut connections" are performed between the input and output of each building block:
        1. If the input and output of the building block have exactly the same dimensions (including number of channels and size), add the input to the output.
        2. If the input and output of the building block have different dimensions (including number of channels and size), add the input to the output after a convolutional layer to make the dimensions match.

        **Args:**
        - **layer_name** (`str`): pass the name of this multi-building-block layer to construct the full name of each weighted convolutional layer.
        - **building_block_type** (`ResNetBlockSmall` | `ResNetBlockLarge`): the type of the building block.
        - **building_block_num** (`int`): the number of building blocks in this multi-building-block layer.
        - **preceding_output_channels** (`int`): the number of channels of preceding output of this entire multi-building-block layer.
        - **input_channels** (`int`): the number of channels of input of this multi-building-block layer.
        - **overall_stride** (`int`): the overall stride of the building blocks. This stride is performed at the 1st building block where other building blocks remain their own overall stride of 1. Inside that building block, this stride is performed at certain convolutional layer in the building block where other convolutional layers remain stride of 1:
            - For `ResNetBlockSmall`, it performs at the 2nd (last) layer.
            - For `ResNetBlockLarge`, it performs at the 2nd (middle) layer.
        - **activation_layer** (`nn.Module`): activation function of each layer (if not `None`), if `None` this layer won't be used. Default `nn.ReLU`.
        - **batch_normalisation** (`bool`): whether to use batch normalisation after the weight convolutional layers. Default `True`, same as what the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html) does.
        - **bias** (`bool`): whether to use bias in the convolutional layer. Default `False`, because batch normalisation are doing the similar thing with bias.

        **Returns:**
        - **layer** (`nn.Sequential`): the constructed layer consisting of multiple building blocks.
        """

        layer = []

        for block_idx in range(building_block_num):
            layer.append(
                building_block_type(
                    outer_layer_name=layer_name,
                    block_idx=block_idx,
                    preceding_output_channels=(
                        preceding_output_channels
                        if block_idx == 0
                        else (
                            input_channels
                            if building_block_type == ResNetBlockSmall
                            else input_channels * 4
                        )
                    ),  # if it's the 1st block in this multi-building-block layer, it should be the number of channels of the preceding output of this entire multi-building-block layer. Otherwise, it should be the number of channels from last building block where the number of channels is 4 times expanded as the input channels for `ResNetBlockLarge` than `ResNetBlockSmall`.
                    input_channels=input_channels,
                    overall_stride=(
                        overall_stride if block_idx == 0 else 1
                    ),  # only perform the overall stride at the 1st block in this multi-building-block layer
                    activation_layer=activation_layer,
                    batch_normalisation=batch_normalisation,
                    bias=bias,
                )
            )

            self.weighted_layer_names += layer[
                -1
            ].weighted_layer_names  # collect the weighted layer names in the blocks and sync to the weighted layer names list in the outer network

        return nn.Sequential(*layer)

    def forward(
        self, input: Tensor, stage: str = None, task_id: int | None = None
    ) -> tuple[Tensor, dict[str, Tensor]]:
        r"""The forward pass for data. It is the same for all tasks.

        **Args:**
        - **input** (`Tensor`): the input tensor from data.

        **Returns:**
        - **output_feature** (`Tensor`): the output feature tensor to be passed into heads. This is the main target of backpropagation.
        - **hidden_features** (`dict[str, Tensor]`): the hidden features (after activation) in each weighted layer. Key (`str`) is the weighted layer name, value (`Tensor`) is the hidden feature tensor. This is used for the continual learning algorithms that need to use the hidden features for various purposes.
        """
        batch_size = input.size(0)
        hidden_features = {}

        x = input

        x = self.conv1(x)
        if self.batch_normalisation:
            x = self.conv_bn1(x)
        if self.activation:
            x = self.conv_activation1(x)
        hidden_features["conv1"] = x

        x = self.maxpool(x)

        for block in self.conv2x:
            x, hidden_features_block = block(x)
            hidden_features.update(hidden_features_block)  # store the hidden feature
        for block in self.conv3x:
            x, hidden_features_block = block(x)
            hidden_features.update(hidden_features_block)  # store the hidden feature
        for block in self.conv4x:
            x, hidden_features_block = block(x)
            hidden_features.update(hidden_features_block)  # store the hidden feature
        for block in self.conv5x:
            x, hidden_features_block = block(x)
            hidden_features.update(hidden_features_block)  # store the hidden feature

        x = self.avepool(x)

        output_feature = x.view(batch_size, -1)  # flatten before going through heads

        return output_feature, hidden_features


class ResNet18(ResNetBase):
    r"""ResNet-18 backbone network.

    This is a smaller architecture proposed in the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html). It consists of 18 weight convolutional layers in total. See Table 1 in the paper for details.
    """

    def __init__(
        self,
        input_channels: int,
        output_dim: int,
        activation_layer: nn.Module | None = nn.ReLU,
        batch_normalisation: bool = True,
        bias: bool = False,
    ) -> None:
        r"""Construct and initialise the ResNet-18 backbone network.

        **Args:**
        - **input_channels** (`int`): the number of channels of input of this building block. Note that convolutional networks require number of input channels instead of dimension.
        - **output_dim** (`int`): the output dimension after flattening at last which connects to CL output heads. Although this is not determined by us but the architecture built before the flattening layer, we still need to provide this to construct the heads.
        - **activation_layer** (`nn.Module`): activation function of each layer (if not `None`), if `None` this layer won't be used. Default `nn.ReLU`.
        - **batch_normalisation** (`bool`): whether to use batch normalisation after the weight convolutional layers. Default `True`, same as what the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html) does.
        - **bias** (`bool`): whether to use bias in the convolutional layer. Default `False`, because batch normalisation are doing the similar thing with bias.
        """
        ResNetBase.__init__(
            self,
            input_channels=input_channels,
            building_block_type=ResNetBlockSmall,  # use the smaller building block for ResNet-18
            building_block_nums=(2, 2, 2, 2),
            building_block_preceding_output_channels=(64, 64, 128, 256),
            building_block_input_channels=(64, 128, 256, 512),
            output_dim=output_dim,
            activation_layer=activation_layer,
            batch_normalisation=batch_normalisation,
            bias=bias,
        )


class ResNet34(ResNetBase):
    r"""ResNet-34 backbone network.

    This is a smaller architecture proposed in the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html). It consists of 34 weight convolutional layers in total. See Table 1 in the paper for details.
    """

    def __init__(
        self,
        input_channels: int,
        output_dim: int,
        activation_layer: nn.Module | None = nn.ReLU,
        batch_normalisation: bool = True,
        bias: bool = False,
    ) -> None:
        r"""Construct and initialise the ResNet-34 backbone network.

        **Args:**
        - **input_channels** (`int`): the number of channels of input of this building block. Note that convolutional networks require number of input channels instead of dimension.
        - **output_dim** (`int`): the output dimension after flattening at last which connects to CL output heads. Although this is not determined by us but the architecture built before the flattening layer, we still need to provide this to construct the heads.
        - **activation_layer** (`nn.Module`): activation function of each layer (if not `None`), if `None` this layer won't be used. Default `nn.ReLU`.
        - **batch_normalisation** (`bool`): whether to use batch normalisation after the weight convolutional layers. Default `True`, same as what the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html) does.
        - **bias** (`bool`): whether to use bias in the convolutional layer. Default `False`, because batch normalisation are doing the similar thing with bias.
        """
        ResNetBase.__init__(
            self,
            input_channels=input_channels,
            building_block_type=ResNetBlockSmall,  # use the smaller building block for ResNet-34
            building_block_nums=(3, 4, 6, 3),
            building_block_preceding_output_channels=(64, 64, 128, 256),
            building_block_input_channels=(64, 128, 256, 512),
            output_dim=output_dim,
            activation_layer=activation_layer,
            batch_normalisation=batch_normalisation,
            bias=bias,
        )


class ResNet50(ResNetBase):
    r"""ResNet-50 backbone network.

    This is a larger architecture proposed in the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html). It consists of 50 weight convolutional layers in total. See Table 1 in the paper for details.
    """

    def __init__(
        self,
        input_channels: int,
        output_dim: int,
        activation_layer: nn.Module | None = nn.ReLU,
        batch_normalisation: bool = True,
        bias: bool = False,
    ) -> None:
        r"""Construct and initialise the ResNet-50 backbone network.

        **Args:**
        - **input_channels** (`int`): the number of channels of input of this building block. Note that convolutional networks require number of input channels instead of dimension.
        - **output_dim** (`int`): the output dimension after flattening at last which connects to CL output heads. Although this is not determined by us but the architecture built before the flattening layer, we still need to provide this to construct the heads.
        - **activation_layer** (`nn.Module`): activation function of each layer (if not `None`), if `None` this layer won't be used. Default `nn.ReLU`.
        - **batch_normalisation** (`bool`): whether to use batch normalisation after the weight convolutional layers. Default `True`, same as what the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html) does.
        - **bias** (`bool`): whether to use bias in the convolutional layer. Default `False`, because batch normalisation are doing the similar thing with bias.
        """
        ResNetBase.__init__(
            self,
            input_channels=input_channels,
            building_block_type=ResNetBlockLarge,  # use the larger building block for ResNet-50
            building_block_nums=(3, 4, 6, 3),
            building_block_preceding_output_channels=(64, 256, 512, 1024),
            building_block_input_channels=(64, 128, 256, 512),
            output_dim=output_dim,
            activation_layer=activation_layer,
            batch_normalisation=batch_normalisation,
            bias=bias,
        )


class ResNet101(ResNetBase):
    r"""ResNet-101 backbone network.

    This is a larger architecture proposed in the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html). It consists of 101 weight convolutional layers in total. See Table 1 in the paper for details.
    """

    def __init__(
        self,
        input_channels: int,
        output_dim: int,
        activation_layer: nn.Module | None = nn.ReLU,
        batch_normalisation: bool = True,
        bias: bool = False,
    ) -> None:
        r"""Construct and initialise the ResNet-101 backbone network.

        **Args:**
        - **input_channels** (`int`): the number of channels of input of this building block. Note that convolutional networks require number of input channels instead of dimension.
        - **output_dim** (`int`): the output dimension after flattening at last which connects to CL output heads. Although this is not determined by us but the architecture built before the flattening layer, we still need to provide this to construct the heads.
        - **activation_layer** (`nn.Module`): activation function of each layer (if not `None`), if `None` this layer won't be used. Default `nn.ReLU`.
        - **batch_normalisation** (`bool`): whether to use batch normalisation after the weight convolutional layers. Default `True`, same as what the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html) does.
        - **bias** (`bool`): whether to use bias in the convolutional layer. Default `False`, because batch normalisation are doing the similar thing with bias.
        """
        ResNetBase.__init__(
            self,
            input_channels=input_channels,
            building_block_type=ResNetBlockLarge,  # use the larger building block for ResNet-101
            building_block_nums=(3, 4, 23, 3),
            building_block_preceding_output_channels=(64, 256, 512, 1024),
            building_block_input_channels=(64, 128, 256, 512),
            output_dim=output_dim,
            activation_layer=activation_layer,
            batch_normalisation=batch_normalisation,
            bias=bias,
        )


class ResNet152(ResNetBase):
    r"""ResNet-152 backbone network.

    This is the largest architecture proposed in the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html). It consists of 152 weight convolutional layers in total. See Table 1 in the paper for details.
    """

    def __init__(
        self,
        input_channels: int,
        output_dim: int,
        activation_layer: nn.Module | None = nn.ReLU,
        batch_normalisation: bool = True,
        bias: bool = False,
    ) -> None:
        r"""Construct and initialise the ResNet-50 backbone network.

        **Args:**
        - **input_channels** (`int`): the number of channels of input of this building block. Note that convolutional networks require number of input channels instead of dimension.
        - **output_dim** (`int`): the output dimension after flattening at last which connects to CL output heads. Although this is not determined by us but the architecture built before the flattening layer, we still need to provide this to construct the heads.
        - **activation_layer** (`nn.Module`): activation function of each layer (if not `None`), if `None` this layer won't be used. Default `nn.ReLU`.
        - **batch_normalisation** (`bool`): whether to use batch normalisation after the weight convolutional layers. Default `True`, same as what the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html) does.
        - **bias** (`bool`): whether to use bias in the convolutional layer. Default `False`, because batch normalisation are doing the similar thing with bias.
        """
        ResNetBase.__init__(
            self,
            input_channels=input_channels,
            building_block_type=ResNetBlockLarge,  # use the larger building block for ResNet-152
            building_block_nums=(3, 8, 36, 3),
            building_block_preceding_output_channels=(64, 256, 512, 1024),
            building_block_input_channels=(64, 128, 256, 512),
            output_dim=output_dim,
            activation_layer=activation_layer,
            batch_normalisation=batch_normalisation,
            bias=bias,
        )


class HATMaskResNetBlockSmall(HATMaskBackbone, ResNetBlockSmall):
    r"""The smaller building block for HAT masked ResNet-18/34.

    It consists of 2 weight convolutional layers, each followed by an activation function. See Table 1 or Figure 5 (left) in the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html).

    Mask is applied to the units which are output channels in each weighted convolutional layer. The mask is generated from the unit-wise task embedding and gate function.
    """

    def __init__(
        self,
        outer_layer_name: str,
        block_idx: int,
        preceding_output_channels: int,
        input_channels: int,
        overall_stride: int,
        gate: str,
        activation_layer: nn.Module | None = nn.ReLU,
        bias: bool = False,
    ) -> None:
        r"""Construct and initialise the smaller building block with task embedding.

        **Args:**
        - **outer_layer_name** (`str`): pass the name of the multi-building-block layer that contains this building block to construct the full name of each weighted convolutional layer.
        - **block_idx** (`int`): the index of the building blocks in the multi-building-block layer to construct the full name of each weighted convolutional layer.
        - **preceding_output_channels** (`int`): the number of channels of preceding output of this particular building block.
        - **input_channels** (`int`): the number of channels of input of this building block.
        - **overall_stride** (`int`): the overall stride of this building block. This stride is performed at 2nd (last) convolutional layer where the 1st convolutional layer remain stride of 1.
        - **gate** (`str`): the type of gate function turning the real value task embeddings into attention masks, should be one of the following:
            - `sigmoid`: the sigmoid function.
        - **activation_layer** (`nn.Module`): activation function of each layer (if not `None`), if `None` this layer won't be used. Default `nn.ReLU`.
        - **bias** (`bool`): whether to use bias in the convolutional layer. Default `False`.
        """
        HATMaskBackbone.__init__(self, output_dim=None, gate=gate)
        ResNetBlockSmall.__init__(
            self,
            outer_layer_name=outer_layer_name,
            block_idx=block_idx,
            preceding_output_channels=preceding_output_channels,
            input_channels=input_channels,
            overall_stride=overall_stride,
            activation_layer=activation_layer,
            bias=bias,
        )
        self.register_hat_mask_module_explicitly(gate=gate)

        # construct the task embedding over the 1st weighted convolutional layer. It is channel-wise
        layer_output_channels = (
            input_channels  # the output channels of the 1st convolutional layer
        )
        self.task_embedding_t[self.full_1st_layer_name] = nn.Embedding(
            num_embeddings=1, embedding_dim=layer_output_channels
        )

        # construct the task embedding over the 2nd weighted convolutional layer. It is channel-wise
        layer_output_channels = (
            input_channels * 1
        )  # the output channels of the 2nd convolutional layer, which is the same as the input channels (without expansion)
        self.task_embedding_t[self.full_2nd_layer_name] = nn.Embedding(
            num_embeddings=1, embedding_dim=layer_output_channels
        )

    def forward(
        self,
        input: Tensor,
        stage: str,
        s_max: float | None = None,
        batch_idx: int | None = None,
        num_batches: int | None = None,
        test_mask: dict[str, Tensor] | None = None,
    ) -> tuple[Tensor, dict[str, Tensor], dict[str, Tensor]]:
        r"""The forward pass for data from task `task_id`. Task-specific mask for `task_id` are applied to the units which are channels in each weighted convolutional layer.

        **Args:**
        - **input** (`Tensor`): The input tensor from data.
        - **stage** (`str`): the stage of the forward pass, should be one of the following:
            1. 'train': training stage.
            2. 'validation': validation stage.
            3. 'test': testing stage.
        - **s_max** (`float` | `None`): the maximum scaling factor in the gate function. Doesn't apply to testing stage. See chapter 2.4 "Hard Attention Training" in [HAT paper](http://proceedings.mlr.press/v80/serra18a).
        - **batch_idx** (`int` | `None`): the current batch index. Applies only to training stage. For other stages, it is default `None`.
        - **num_batches** (`int` | `None`): the total number of batches. Applies only to training stage. For other stages, it is default `None`.
        - **test_mask** (`dict[str, Tensor]` | `None`): the binary mask used for test. Applies only to testing stage. For other stages, it is default `None`.

        **Returns:**
        - **output_feature** (`Tensor`): the output feature maps.
        - **mask** (`dict[str, Tensor]`): the mask for the current task. Key (`str`) is layer name, value (`Tensor`) is the mask tensor. The mask tensor has size (number of units).
        - **hidden_features** (`dict[str, Tensor]`): the hidden features (after activation) in each weighted layer. Key (`str`) is the weighted layer name, value (`Tensor`) is the hidden feature tensor. This is used for the continual learning algorithms that need to use the hidden features for various purposes. Although HAT algorithm does not need this, it is still provided for API consistence for other HAT-based algorithms inherited this `forward()` method of `HAT` class.
        """
        hidden_features = {}

        # get the mask for the current task from the task embedding in this stage
        mask = self.get_mask(
            stage=stage,
            s_max=s_max,
            batch_idx=batch_idx,
            num_batches=num_batches,
            test_mask=test_mask,
        )

        identity = (
            self.identity_downsample(input)
            if self.identity_downsample is not None
            else input
        )  # remember the identity of input for the shortcut connection. Perform downsampling if its dimension doesn't match the output's

        x = input
        x = self.conv1(x)  # weighted convolutional layer first
        x = x * (
            mask[self.full_1st_layer_name].view(1, -1, 1, 1)
        )  # apply the mask to the 1st convolutional layer. Broadcast the dimension of mask to match the input
        if self.activation:
            x = self.conv_activation1(x)  # activation function third
        hidden_features[self.full_1st_layer_name] = x  # store the hidden feature

        x = self.conv2(x)  # weighted convolutional layer first
        x = x + identity
        x = x * (
            mask[self.full_2nd_layer_name].view(1, -1, 1, 1)
        )  # apply the mask to the 2nd convolutional layer after the shortcut connection. Broadcast the dimension of mask to match the input
        if self.activation:
            x = self.conv_activation2(x)  # activation after the shortcut connection
        hidden_features[self.full_2nd_layer_name] = x  # store the hidden feature

        output_feature = x

        return output_feature, mask, hidden_features


class HATMaskResNetBlockLarge(HATMaskBackbone, ResNetBlockLarge):
    r"""The larger building block for ResNet-50/101/152. It is referred to "bottleneck" building block in the ResNet paper.

    It consists of 3 weight convolutional layers, each followed by an activation function. See Table 1 or Figure 5 (right) in the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html).

    Mask is applied to the units which are output channels in each weighted convolutional layer. The mask is generated from the unit-wise task embedding and gate function.
    """

    def __init__(
        self,
        outer_layer_name: str,
        block_idx: int,
        preceding_output_channels: int,
        input_channels: int,
        overall_stride: int,
        gate: str,
        activation_layer: nn.Module | None = nn.ReLU,
        bias: bool = False,
    ) -> None:
        r"""Construct and initialise the larger building block with task embedding.

        **Args:**
        - **outer_layer_name** (`str`): pass the name of the multi-building-block layer that contains this building block to construct the full name of each weighted convolutional layer.
        - **block_idx** (`int`): the index of the building blocks in the multi-building-block layer to construct the full name of each weighted convolutional layer.
        - **preceding_output_channels** (`int`): the number of channels of preceding output of this particular building block.
        - **input_channels** (`int`): the number of channels of input of this building block.
        - **overall_stride** (`int`): the overall stride of this building block. This stride is performed at 2nd (middle) convolutional layer where 1st and 3rd convolutional layers remain stride of 1.
        - **gate** (`str`): the type of gate function turning the real value task embeddings into attention masks, should be one of the following:
            - `sigmoid`: the sigmoid function.
        - **activation_layer** (`nn.Module`): activation function of each layer (if not `None`), if `None` this layer won't be used. Default `nn.ReLU`.
        - **bias** (`bool`): whether to use bias in the convolutional layer. Default `False`.
        """
        HATMaskBackbone.__init__(self, output_dim=None, gate=gate)
        ResNetBlockLarge.__init__(
            self,
            outer_layer_name=outer_layer_name,
            block_idx=block_idx,
            preceding_output_channels=preceding_output_channels,
            input_channels=input_channels,
            overall_stride=overall_stride,
            activation_layer=activation_layer,
            bias=bias,
        )
        self.register_hat_mask_module_explicitly(gate=gate)

        # construct the task embedding over the 1st weighted convolutional layer. It is channel-wise
        layer_output_channels = (
            input_channels  # the output channels of the 1st convolutional layer
        )
        self.task_embedding_t[self.full_1st_layer_name] = nn.Embedding(
            num_embeddings=1, embedding_dim=layer_output_channels
        )

        # construct the task embedding over the 2nd weighted convolutional layer. It is channel-wise
        layer_output_channels = (
            input_channels * 1
        )  # the output channels of the 2nd convolutional layer, which is the same as the input channels (without expansion)
        self.task_embedding_t[self.full_2nd_layer_name] = nn.Embedding(
            num_embeddings=1, embedding_dim=layer_output_channels
        )

        # construct the task embedding over the 3rd weighted convolutional layer. It is channel-wise
        layer_output_channels = (
            input_channels
            * 4  # the output channels of the 2nd convolutional layer, which is 4 times expanded as the input channels
        )
        self.task_embedding_t[self.full_3rd_layer_name] = nn.Embedding(
            num_embeddings=1, embedding_dim=layer_output_channels
        )

    def forward(
        self,
        input: Tensor,
        stage: str,
        s_max: float | None = None,
        batch_idx: int | None = None,
        num_batches: int | None = None,
        test_mask: dict[str, Tensor] | None = None,
    ) -> tuple[Tensor, dict[str, Tensor], dict[str, Tensor]]:
        r"""The forward pass for data from task `task_id`. Task-specific mask for `task_id` are applied to the units which are channels in each weighted convolutional layer.

        **Args:**
        - **input** (`Tensor`): The input tensor from data.
        - **stage** (`str`): the stage of the forward pass, should be one of the following:
            1. 'train': training stage.
            2. 'validation': validation stage.
            3. 'test': testing stage.
        - **s_max** (`float` | `None`): the maximum scaling factor in the gate function. Doesn't apply to testing stage. See chapter 2.4 "Hard Attention Training" in [HAT paper](http://proceedings.mlr.press/v80/serra18a).
        - **batch_idx** (`int` | `None`): the current batch index. Applies only to training stage. For other stages, it is default `None`.
        - **num_batches** (`int` | `None`): the total number of batches. Applies only to training stage. For other stages, it is default `None`.
        - **test_mask** (`dict[str, Tensor]` | `None`): the binary mask used for test. Applies only to testing stage. For other stages, it is default `None`.

        **Returns:**
        - **output_feature** (`Tensor`): the output feature maps.
        - **mask** (`dict[str, Tensor]`): the mask for the current task. Key (`str`) is layer name, value (`Tensor`) is the mask tensor. The mask tensor has size (number of units).
        - **hidden_features** (`dict[str, Tensor]`): the hidden features (after activation) in each weighted layer. Key (`str`) is the weighted layer name, value (`Tensor`) is the hidden feature tensor. This is used for the continual learning algorithms that need to use the hidden features for various purposes. Although HAT algorithm does not need this, it is still provided for API consistence for other HAT-based algorithms inherited this `forward()` method of `HAT` class.
        """
        hidden_features = {}

        # get the mask for the current task from the task embedding in this stage
        mask = self.get_mask(
            stage=stage,
            s_max=s_max,
            batch_idx=batch_idx,
            num_batches=num_batches,
            test_mask=test_mask,
        )

        identity = (
            self.identity_downsample(input)
            if self.identity_downsample is not None
            else input
        )  # remember the identity of input for the shortcut connection. Perform downsampling if its dimension doesn't match the output's

        x = input
        x = self.conv1(x)  # weighted convolutional layer first
        x = x * (
            mask[self.full_1st_layer_name].view(1, -1, 1, 1)
        )  # apply the mask to the 1st convolutional layer. Broadcast the dimension of mask to match the input
        if self.activation:
            x = self.conv_activation1(x)  # activation function third
        hidden_features[self.full_1st_layer_name] = x  # store the hidden feature

        x = self.conv2(x)  # weighted convolutional layer first
        x = x * (
            mask[self.full_2nd_layer_name].view(1, -1, 1, 1)
        )  # apply the mask to the 2nd convolutional layer. Broadcast the dimension of mask to match the input
        if self.activation:
            x = self.conv_activation2(x)  # activation function third
        hidden_features[self.full_2nd_layer_name] = x  # store the hidden feature

        x = self.conv3(x)  # weighted convolutional layer first
        x = x + identity
        x = x * (
            mask[self.full_3rd_layer_name].view(1, -1, 1, 1)
        )  # apply the mask to the 3rd convolutional layer after the shortcut connection. Broadcast the dimension of mask to match the input
        if self.activation:
            x = self.activation3(x)  # activation after the shortcut connection
        hidden_features[self.full_3rd_layer_name] = x  # store the hidden feature

        output_feature = x

        return output_feature, mask, hidden_features


class HATMaskResNetBase(ResNetBase, HATMaskBackbone):
    r"""The base class of HAT masked [residual network (ResNet)](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html).

    [HAT (Hard Attention to the Task, 2018)](http://proceedings.mlr.press/v80/serra18a) is an architecture-based continual learning approach that uses learnable hard attention masks to select the task-specific parameters.

    ResNet is a convolutional network architecture, which has 1st convolutional parameter layer and a maxpooling layer, connecting to 4 convolutional layers which contains multiple convolutional parameter layer. Each layer of the 4 are constructed from basic building blocks which are either small (`ResNetBlockSmall`) or large (`ResNetBlockLarge`). Each building block contains several convolutional parameter layers. The building blocks are connected by a skip connection which is a direct connection from the input of the block to the output of the block, and this is why it's called residual (find "shortcut connections" in the paper for more details). After the 5th convolutional layer, there are average pooling layer and a fully connected layer which connects to the CL output heads.

    Mask is applied to the units which are output channels in each weighted convolutional layer. The mask is generated from the unit-wise task embedding and gate function.
    """

    def __init__(
        self,
        input_channels: int,
        building_block_type: HATMaskResNetBlockSmall | HATMaskResNetBlockLarge,
        building_block_nums: tuple[int, int, int, int],
        building_block_preceding_output_channels: tuple[int, int, int, int],
        building_block_input_channels: tuple[int, int, int, int],
        output_dim: int,
        gate: str,
        activation_layer: nn.Module | None = nn.ReLU,
        bias: bool = False,
    ) -> None:
        r"""Construct and initialise the HAT masked ResNet backbone network with task embedding. Note that batch normalisation is incompatible with HAT mechanism.

        **Args:**
        - **input_channels** (`int`): the number of channels of input. Image data are kept channels when going in ResNet. Note that convolutional networks require number of input channels instead of dimension.
        - **building_block_type** (`HATMaskResNetBlockSmall` | `HATMaskResNetBlockLarge`): the type of building block used in the ResNet.
        - **building_block_nums** (`tuple[int, int, int, int]`): the number of building blocks in the 2-5 convolutional layer correspondingly.
        - **building_block_preceding_output_channels** (`tuple[int, int, int, int]`): the number of channels of preceding output of each building block in the 2-5 convolutional layer correspondingly.
        - **building_block_input_channels** (`tuple[int, int, int, int]`): the number of channels of input of each building block in the 2-5 convolutional layer correspondingly.
        - **output_dim** (`int`): the output dimension after flattening at last which connects to CL output heads. Although this is not determined by us but the architecture built before the flattening layer, we still need to provide this to construct the heads.
        - **gate** (`str`): the type of gate function turning the real value task embeddings into attention masks, should be one of the following:
            - `sigmoid`: the sigmoid function.
        - **activation_layer** (`nn.Module`): activation function of each layer (if not `None`), if `None` this layer won't be used. Default `nn.ReLU`.
        - **bias** (`bool`): whether to use bias in the convolutional layer. Default `False`, because batch normalisation are doing the similar thing with bias.
        """
        # init from both inherited classes
        HATMaskBackbone.__init__(self, output_dim=output_dim, gate=gate)
        ResNetBase.__init__(
            self,
            input_channels=input_channels,
            building_block_type=building_block_type,
            building_block_nums=building_block_nums,
            building_block_preceding_output_channels=building_block_preceding_output_channels,
            building_block_input_channels=building_block_input_channels,
            output_dim=output_dim,
            activation_layer=activation_layer,
            batch_normalisation=False,  # batch normalisation is incompatible with HAT mechanism
            bias=bias,
        )
        self.register_hat_mask_module_explicitly(
            gate=gate
        )  # register all `nn.Module`s for HATMaskBackbone explicitly because the second `__init__()` wipes out them inited by the first `__init__()`
        self.update_multiple_blocks_task_embedding()

        # construct the task embedding over the 1st weighted convolutional layers. It is channel-wise
        layer_output_channels = 64  # the output channels of the 1st convolutional layer
        self.task_embedding_t["conv1"] = nn.Embedding(
            num_embeddings=1, embedding_dim=layer_output_channels
        )

    def _multiple_blocks(
        self,
        layer_name: str,
        building_block_type: HATMaskResNetBlockSmall | HATMaskResNetBlockLarge,
        building_block_num: int,
        preceding_output_channels: int,
        input_channels: int,
        overall_stride: int,
        activation_layer: nn.Module | None = nn.ReLU,
        batch_normalisation: bool = False,
        bias: bool = False,
    ) -> None:
        r"""Construct a layer consisting of multiple building blocks with task embedding. It's used to construct the 2-5 convolutional layers of the HAT masked ResNet.

        The "shortcut connections" are performed between the input and output of each building block:
        1. If the input and output of the building block have exactly the same dimensions (including number of channels and size), add the input to the output.
        2. If the input and output of the building block have different dimensions (including number of channels and size), add the input to the output after a convolutional layer to make the dimensions match.

        **Args:**
        - **layer_name** (`str`): pass the name of this multi-building-block layer to construct the full name of each weighted convolutional layer.
        - **building_block_type** (`HATMaskResNetBlockSmall` | `HATMaskResNetBlockLarge`): the type of the building block.
        - **building_block_num** (`int`): the number of building blocks in this multi-building-block layer.
        - **preceding_output_channels** (`int`): the number of channels of preceding output of this entire multi-building-block layer.
        - **input_channels** (`int`): the number of channels of input of this multi-building-block layer.
        - **overall_stride** (`int`): the overall stride of the building blocks. This stride is performed at the 1st building block where other building blocks remain their own overall stride of 1. Inside that building block, this stride is performed at certain convolutional layer in the building block where other convolutional layers remain stride of 1:
            - For `ResNetBlockSmall`, it performs at the 2nd (last) layer.
            - For `ResNetBlockLarge`, it performs at the 2nd (middle) layer.
        - **activation_layer** (`nn.Module`): activation function of each layer (if not `None`), if `None` this layer won't be used. Default `nn.ReLU`.
        - **batch_normalisation** (`bool`): whether to use batch normalisation after the weight convolutional layers. In HATMaskResNet, batch normalisation is incompatible with HAT mechanism and shoule be always set `False`. We include this argument for compatibility with the original ResNet API.
        - **bias** (`bool`): whether to use bias in the convolutional layer. Default `False`.

        **Returns:**
        - **layer** (`nn.Sequential`): the constructed layer consisting of multiple building blocks.
        """

        layer = []

        for block_idx in range(building_block_num):
            layer.append(
                building_block_type(
                    outer_layer_name=layer_name,
                    block_idx=block_idx,
                    preceding_output_channels=(
                        preceding_output_channels
                        if block_idx == 0
                        else (
                            input_channels
                            if building_block_type == HATMaskResNetBlockSmall
                            else input_channels * 4
                        )
                    ),  # if it's the 1st block in this multi-building-block layer, it should be the number of channels of the preceding output of this entire multi-building-block layer. Otherwise, it should be the number of channels from last building block where the number of channels is 4 times of the input channels for `ResNetBlockLarge` than `ResNetBlockSmall`.
                    input_channels=input_channels,
                    overall_stride=(
                        overall_stride if block_idx == 0 else 1
                    ),  # only perform the overall stride at the 1st block in this multi-building-block layer
                    gate=self.gate,
                    # no batch normalisation in HAT masked blocks
                    activation_layer=activation_layer,
                    bias=bias,
                )
            )

            self.weighted_layer_names += layer[
                -1
            ].weighted_layer_names  # collect the weighted layer names in the blocks and sync to the weighted layer names list in the outer network

        return nn.Sequential(*layer)

    def update_multiple_blocks_task_embedding(self) -> None:
        r"""Collect the task embeddings in the multiple building blocks (2-5 convolutional layers) and sync to the weighted layer names list in the outer network.

        This should only be called explicitly after the `__init__()` method, just because task embedding as `nn.Module` instance was wiped out at the beginning of it.
        """
        for block in self.conv2x:
            self.task_embedding_t.update(block.task_embedding_t)
        for block in self.conv3x:
            self.task_embedding_t.update(block.task_embedding_t)
        for block in self.conv4x:
            self.task_embedding_t.update(block.task_embedding_t)
        for block in self.conv5x:
            self.task_embedding_t.update(block.task_embedding_t)

    def forward(
        self,
        input: Tensor,
        stage: str,
        s_max: float | None = None,
        batch_idx: int | None = None,
        num_batches: int | None = None,
        test_mask: dict[str, Tensor] | None = None,
    ) -> tuple[Tensor, dict[str, Tensor], dict[str, Tensor]]:
        r"""The forward pass for data from task `task_id`. Task-specific mask for `task_id` are applied to the units which are channels in each weighted convolutional layer.

        **Args:**
        - **input** (`Tensor`): the input tensor from data.
        - **stage** (`str`): the stage of the forward pass, should be one of the following:
            1. 'train': training stage.
            2. 'validation': validation stage.
            3. 'test': testing stage.
        - **s_max** (`float` | `None`): the maximum scaling factor in the gate function. Doesn't apply to testing stage. See chapter 2.4 "Hard Attention Training" in [HAT paper](http://proceedings.mlr.press/v80/serra18a).
        - **batch_idx** (`int` | `None`): the current batch index. Applies only to training stage. For other stages, it is default `None`.
        - **num_batches** (`int` | `None`): the total number of batches. Applies only to training stage. For other stages, it is default `None`.
        - **test_mask** (`dict[str, Tensor]` | `None`): the binary mask used for test. Applies only to testing stage. For other stages, it is default `None`.

        **Returns:**
        - **output_feature** (`Tensor`): the output feature tensor to be passed to the heads.
        - **mask** (`dict[str, Tensor]`): the mask for the current task. Key (`str`) is layer name, value (`Tensor`) is the mask tensor. The mask tensor has size (number of units).
        - **hidden_features** (`dict[str, Tensor]`): the hidden features (after activation) in each weighted layer. Key (`str`) is the weighted layer name, value (`Tensor`) is the hidden feature tensor. This is used for the continual learning algorithms that need to use the hidden features for various purposes. Although HAT algorithm does not need this, it is still provided for API consistence for other HAT-based algorithms inherited this `forward()` method of `HAT` class.
        """
        batch_size = input.size(0)
        hidden_features = {}

        # get the mask for the current task from the task embedding in this stage
        mask = self.get_mask(
            stage=stage,
            s_max=s_max,
            batch_idx=batch_idx,
            num_batches=num_batches,
            test_mask=test_mask,
        )

        x = input

        x = self.conv1(x)

        x = x * (
            mask["conv1"].view(1, -1, 1, 1)
        )  # apply the mask to the 1st convolutional layer. Broadcast the dimension of mask to match the input
        if self.activation:
            x = self.conv_activation1(x)
        hidden_features["conv1"] = x

        x = self.maxpool(x)

        for block in self.conv2x:
            x, _, hidden_features_block = block(
                x,
                stage=stage,
                s_max=s_max,
                batch_idx=batch_idx,
                num_batches=num_batches,
                test_mask=test_mask,
            )
            hidden_features.update(hidden_features_block)  # store the hidden feature
        for block in self.conv3x:
            x, _, hidden_features_block = block(
                x,
                stage=stage,
                s_max=s_max,
                batch_idx=batch_idx,
                num_batches=num_batches,
                test_mask=test_mask,
            )
            hidden_features.update(hidden_features_block)  # store the hidden feature
        for block in self.conv4x:
            x, _, hidden_features_block = block(
                x,
                stage=stage,
                s_max=s_max,
                batch_idx=batch_idx,
                num_batches=num_batches,
                test_mask=test_mask,
            )
            hidden_features.update(hidden_features_block)  # store the hidden feature
        for block in self.conv5x:
            x, _, hidden_features_block = block(
                x,
                stage=stage,
                s_max=s_max,
                batch_idx=batch_idx,
                num_batches=num_batches,
                test_mask=test_mask,
            )
            hidden_features.update(hidden_features_block)  # store the hidden feature

        x = self.avepool(x)

        output_feature = x.view(batch_size, -1)  # flatten before going through heads

        return output_feature, mask, hidden_features


class HATMaskResNet18(HATMaskResNetBase):
    r"""HAT masked ResNet-18 backbone network.

    [HAT (Hard Attention to the Task, 2018)](http://proceedings.mlr.press/v80/serra18a) is an architecture-based continual learning approach that uses learnable hard attention masks to select the task-specific parameters.

    ResNet-18 is a smaller architecture proposed in the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html). It consists of 18 weight convolutional layers in total. See Table 1 in the paper for details.

    Mask is applied to the units which are output channels in each weighted convolutional layer. The mask is generated from the unit-wise task embedding and gate function.
    """

    def __init__(
        self,
        input_channels: int,
        output_dim: int,
        gate: str,
        activation_layer: nn.Module | None = nn.ReLU,
        bias: bool = False,
    ) -> None:
        r"""Construct and initialise the ResNet-18 backbone network with task embedding. Note that batch normalisation is incompatible with HAT mechanism.

        **Args:**
        - **input_channels** (`int`): the number of channels of input of this building block. Note that convolutional networks require number of input channels instead of dimension.
        - **output_dim** (`int`): the output dimension after flattening at last which connects to CL output heads. Although this is not determined by us but the architecture built before the flattening layer, we still need to provide this to construct the heads.
        - **gate** (`str`): the type of gate function turning the real value task embeddings into attention masks, should be one of the following:
            - `sigmoid`: the sigmoid function.
        - **activation_layer** (`nn.Module`): activation function of each layer (if not `None`), if `None` this layer won't be used. Default `nn.ReLU`.
        - **bias** (`bool`): whether to use bias in the convolutional layer. Default `False`.
        """
        HATMaskResNetBase.__init__(
            self,
            input_channels=input_channels,
            building_block_type=HATMaskResNetBlockSmall,  # use the smaller building block for ResNet-18
            building_block_nums=(2, 2, 2, 2),
            building_block_preceding_output_channels=(64, 64, 128, 256),
            building_block_input_channels=(64, 128, 256, 512),
            output_dim=output_dim,
            gate=gate,
            activation_layer=activation_layer,
            bias=bias,
        )


class HATMaskResNet34(HATMaskResNetBase):
    r"""HAT masked ResNet-34 backbone network.

    [HAT (Hard Attention to the Task, 2018)](http://proceedings.mlr.press/v80/serra18a) is an architecture-based continual learning approach that uses learnable hard attention masks to select the task-specific parameters.

    ResNet-34 is a smaller architecture proposed in the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html). It consists of 34 weight convolutional layers in total. See Table 1 in the paper for details.

    Mask is applied to the units which are output channels in each weighted convolutional layer. The mask is generated from the unit-wise task embedding and gate function.
    """

    def __init__(
        self,
        input_channels: int,
        output_dim: int,
        gate: str,
        activation_layer: nn.Module | None = nn.ReLU,
        bias: bool = False,
    ) -> None:
        r"""Construct and initialise the ResNet-34 backbone network with task embedding. Note that batch normalisation is incompatible with HAT mechanism.

        **Args:**
        - **input_channels** (`int`): the number of channels of input of this building block. Note that convolutional networks require number of input channels instead of dimension.
        - **output_dim** (`int`): the output dimension after flattening at last which connects to CL output heads. Although this is not determined by us but the architecture built before the flattening layer, we still need to provide this to construct the heads.
        - **gate** (`str`): the type of gate function turning the real value task embeddings into attention masks, should be one of the following:
            - `sigmoid`: the sigmoid function.
        - **activation_layer** (`nn.Module`): activation function of each layer (if not `None`), if `None` this layer won't be used. Default `nn.ReLU`.
        - **bias** (`bool`): whether to use bias in the convolutional layer. Default `False`.
        """
        HATMaskResNetBase.__init__(
            self,
            input_channels=input_channels,
            building_block_type=HATMaskResNetBlockSmall,  # use the smaller building block for ResNet-34
            building_block_nums=(3, 4, 6, 3),
            building_block_preceding_output_channels=(64, 64, 128, 256),
            building_block_input_channels=(64, 128, 256, 512),
            output_dim=output_dim,
            gate=gate,
            activation_layer=activation_layer,
            bias=bias,
        )


class HATMaskResNet50(HATMaskResNetBase):
    r"""HAT masked ResNet-50 backbone network.

    [HAT (Hard Attention to the Task, 2018)](http://proceedings.mlr.press/v80/serra18a) is an architecture-based continual learning approach that uses learnable hard attention masks to select the task-specific parameters.

    ResNet-50 is a larger architecture proposed in the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html). It consists of 50 weight convolutional layers in total. See Table 1 in the paper for details.

    Mask is applied to the units which are output channels in each weighted convolutional layer. The mask is generated from the unit-wise task embedding and gate function.
    """

    def __init__(
        self,
        input_channels: int,
        output_dim: int,
        gate: str,
        activation_layer: nn.Module | None = nn.ReLU,
        bias: bool = False,
    ) -> None:
        r"""Construct and initialise the ResNet-50 backbone network with task embedding. Note that batch normalisation is incompatible with HAT mechanism.

        **Args:**
        - **input_channels** (`int`): the number of channels of input of this building block. Note that convolutional networks require number of input channels instead of dimension.
        - **output_dim** (`int`): the output dimension after flattening at last which connects to CL output heads. Although this is not determined by us but the architecture built before the flattening layer, we still need to provide this to construct the heads.
        - **gate** (`str`): the type of gate function turning the real value task embeddings into attention masks, should be one of the following:
            - `sigmoid`: the sigmoid function.
        - **activation_layer** (`nn.Module`): activation function of each layer (if not `None`), if `None` this layer won't be used. Default `nn.ReLU`.
        - **bias** (`bool`): whether to use bias in the convolutional layer. Default `False`.
        """
        HATMaskResNetBase.__init__(
            self,
            input_channels=input_channels,
            building_block_type=HATMaskResNetBlockLarge,  # use the smaller building block for ResNet-50
            building_block_nums=(3, 4, 6, 3),
            building_block_preceding_output_channels=(64, 256, 512, 1024),
            building_block_input_channels=(64, 128, 256, 512),
            output_dim=output_dim,
            gate=gate,
            activation_layer=activation_layer,
            bias=bias,
        )


class HATMaskResNet101(HATMaskResNetBase):
    r"""HAT masked ResNet-101 backbone network.

    [HAT (Hard Attention to the Task, 2018)](http://proceedings.mlr.press/v80/serra18a) is an architecture-based continual learning approach that uses learnable hard attention masks to select the task-specific parameters.

    ResNet-101 is a larger architecture proposed in the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html). It consists of 101 weight convolutional layers in total. See Table 1 in the paper for details.

    Mask is applied to the units which are output channels in each weighted convolutional layer. The mask is generated from the unit-wise task embedding and gate function.
    """

    def __init__(
        self,
        input_channels: int,
        output_dim: int,
        gate: str,
        activation_layer: nn.Module | None = nn.ReLU,
        bias: bool = False,
    ) -> None:
        r"""Construct and initialise the ResNet-101 backbone network with task embedding. Note that batch normalisation is incompatible with HAT mechanism.

        **Args:**
        - **input_channels** (`int`): the number of channels of input of this building block. Note that convolutional networks require number of input channels instead of dimension.
        - **output_dim** (`int`): the output dimension after flattening at last which connects to CL output heads. Although this is not determined by us but the architecture built before the flattening layer, we still need to provide this to construct the heads.
        - **gate** (`str`): the type of gate function turning the real value task embeddings into attention masks, should be one of the following:
            - `sigmoid`: the sigmoid function.
        - **activation_layer** (`nn.Module`): activation function of each layer (if not `None`), if `None` this layer won't be used. Default `nn.ReLU`.
        - **bias** (`bool`): whether to use bias in the convolutional layer. Default `False`.
        """
        HATMaskResNetBase.__init__(
            self,
            input_channels=input_channels,
            building_block_type=HATMaskResNetBlockLarge,  # use the smaller building block for ResNet-18
            building_block_nums=(3, 4, 23, 3),
            building_block_preceding_output_channels=(64, 256, 512, 1024),
            building_block_input_channels=(64, 128, 256, 512),
            output_dim=output_dim,
            gate=gate,
            activation_layer=activation_layer,
            bias=bias,
        )


class HATMaskResNet152(HATMaskResNetBase):
    r"""HAT masked ResNet-152 backbone network.

    [HAT (Hard Attention to the Task, 2018)](http://proceedings.mlr.press/v80/serra18a) is an architecture-based continual learning approach that uses learnable hard attention masks to select the task-specific parameters.

    ResNet-152 is the largest architecture proposed in the [original ResNet paper](https://www.cv-foundation.org/openaccess/content_cvpr_2016/html/He_Deep_Residual_Learning_CVPR_2016_paper.html). It consists of 152 weight convolutional layers in total. See Table 1 in the paper for details.

    Mask is applied to the units which are output channels in each weighted convolutional layer. The mask is generated from the unit-wise task embedding and gate function.
    """

    def __init__(
        self,
        input_channels: int,
        output_dim: int,
        gate: str,
        activation_layer: nn.Module | None = nn.ReLU,
        bias: bool = False,
    ) -> None:
        r"""Construct and initialise the ResNet-152 backbone network with task embedding. Note that batch normalisation is incompatible with HAT mechanism.

        **Args:**
        - **input_channels** (`int`): the number of channels of input of this building block. Note that convolutional networks require number of input channels instead of dimension.
        - **output_dim** (`int`): the output dimension after flattening at last which connects to CL output heads. Although this is not determined by us but the architecture built before the flattening layer, we still need to provide this to construct the heads.
        - **gate** (`str`): the type of gate function turning the real value task embeddings into attention masks, should be one of the following:
            - `sigmoid`: the sigmoid function.
        - **activation_layer** (`nn.Module`): activation function of each layer (if not `None`), if `None` this layer won't be used. Default `nn.ReLU`.
        - **bias** (`bool`): whether to use bias in the convolutional layer. Default `False`.
        """
        HATMaskResNetBase.__init__(
            self,
            input_channels=input_channels,
            building_block_type=HATMaskResNetBlockLarge,  # use the smaller building block for ResNet-152
            building_block_nums=(3, 8, 36, 3),
            building_block_preceding_output_channels=(64, 256, 512, 1024),
            building_block_input_channels=(64, 128, 256, 512),
            output_dim=output_dim,
            gate=gate,
            activation_layer=activation_layer,
            bias=bias,
        )



r"""

# Callbacks

This submodule provides **callbacks** that can be used in CLArena. 

Please note that this is an API documantation. Please refer to the main documentation pages for more information about the callbacks and how to configure and implement them:

- [**Configure Callbacks**](https://pengxiang-wang.com/projects/continual-learning-arena/docs/configure-your-experiment/callbacks)
- [**Implement Your Callbacks**](https://pengxiang-wang.com/projects/continual-learning-arena/docs/implement-your-cl-modules/callback)

The callbacks are implemented as subclasses of `lightning.Callback`.

"""

from .cl_metrics import CLMetricsCallback
from .cl_rich_progress_bar import CLRichProgressBar
from .hat_metrics import HATMetricsCallback
from .pylogger import PyloggerCallback
from .save_first_batch_images import SaveFirstBatchImagesCallback

__all__ = [
    "cl_rich_progress_bar",
    "save_first_batch_images",
    "cl_metrics",
    "pylogger",
    "hat_metrics",
]



r"""

# Welcome to CLArena

**CLArena (Continual Learning Arena)** is a open-source Python package for Continual Learning (CL) research. In this package, we provide a integrated environment and various APIs to conduct CL experiments for research purposes, as well as implemented CL algorithms and datasets that you can give it a spin immediately.

Please note that this is an API documantation providing detailed information about the available classes, functions, and modules in CLArena. Please refer to the main documentation and my beginners' guide to continual learning for more intuitive tutorials, examples, and guides on how to use CLArena:

- [**Main Documentation**](https://pengxiang-wang.com/projects/continual-learning-arena)
- [**A Beginners' Guide to Continual Learning**](https://pengxiang-wang.com/posts/continual-learning-beginners-guide)

We provide various components of continual learning system in the submodules:

- `clarena.cl_datasets`: Continual learning datasets.
- `clarena.backbones`: Neural network architectures used as backbones for CL algorithms.
- `clarena.cl_heads`: Multi-head classifiers for continual learning outputs. Task-Incremental Learning (TIL) head and Class-Incremental Learning (CIL) head are included.
- `clarena.cl_algorithms`: Implementation of various continual learning algorithms.
- `clarena.callbacks`: Extra actions added in the continual learning process.
- `utils`: Utility functions for continual learning experiments.

As well as the base class in the outmost directory of the package:

- `CLExperiment`: The base class for continual learning experiments.

"""

from .base import CLExperiment

__all__ = [
    "CLExperiment",
    "cl_datasets",
    "backbones",
    "cl_heads",
    "cl_algorithms",
    "callbacks",
    "utils",
]



r"""

# Continual Learning Heads

This submodule provides the **continual learning heads** in CLArena. 

There are two types of heads in CLArena: `HeadsTIL` and `HeadsCIL`, corresponding to two CL paradigms respectively: Task-Incremental Learning (TIL) and Class-Incremental Learning (CIL). 

Please note that this is an API documantation. Please refer to the main documentation pages for more information about the heads.

- [**Configure CL Paradigm in Experiment Index Config**](https://pengxiang-wang.com/projects/continual-learning-arena/docs/configure-your-experiment/experiment-index-config)
- [**A Beginners' Guide to Continual Learning (Multi-head Classifier)](https://pengxiang-wang.com/posts/continual-learning-beginners-guide#sec-CL-classification)

"""

from .heads_cil import HeadsCIL
from .heads_til import HeadsTIL

__all__ = ["HeadsTIL", "HeadsCIL"]



"""The submodule in `utils` for plotting utils."""

__all__ = ["MeanMetricBatch", "HATNetworkCapacity"]

from typing import Any

import torch
from torch import Tensor
from torchmetrics.aggregation import BaseAggregator


class MeanMetricBatch(BaseAggregator):
    r"""A TorchMetrics metric to calculate the mean of metrics across data batches.

    This is used for accumulated metrics in deep learning. See [here](https://pengxiang-wang.com/posts/continual-learning-metrics.html#nte-accumulate) for more details.
    """

    def __init__(self, nan_strategy: str | float = "error", **kwargs: Any) -> None:
        r"""Initialise the metric. Add state variables."""
        BaseAggregator.__init__(
            self,
            "sum",
            torch.tensor(0.0, dtype=torch.get_default_dtype()),
            nan_strategy,
            state_name="sum",
            **kwargs,
        )

        self.sum: Tensor
        r"""State variable created by `super().__init__()` to store the sum of the metric values till this batch."""

        self.add_state(
            "num",
            default=torch.tensor(0, dtype=torch.get_default_dtype()),
            dist_reduce_fx="sum",
        )
        self.num: Tensor
        r"""State variable created by `add_state()` to store the number of the data till this batch."""

    def update(self, value: torch.Tensor, batch_size: int) -> None:
        r"""Update and accumulate the sum of metric value and num of the data till this batch from the batch.

        **Args:**
        - **val** (`torch.Tensor`): the metric value of the batch to update the sum.
        - **batch_size** (`int`): the value to update the num, which is the batch size.
        """

        value = torch.as_tensor(value, dtype=self.dtype, device=self.device)
        batch_size = torch.as_tensor(batch_size, dtype=self.dtype, device=self.device)

        self.sum += value * batch_size
        self.num += batch_size

    def compute(self) -> Tensor:
        r"""Compute this mean metric value till this batch.

        **Returns:**
        - **mean** (`Tensor`): the calculated mean result.
        """
        return self.sum / self.num


class HATNetworkCapacity(BaseAggregator):
    r"""A torchmetrics metric to calculate the network capacity of HAT (Hard Attention to the Task) algorithm.

    Network capacity is defined as the average adjustment rate over all paramaters. See chapter 4.1 in [AdaHAT paper](https://link.springer.com/chapter/10.1007/978-3-031-70352-2_9).
    """

    def __init__(self, nan_strategy: str | float = "error", **kwargs: Any) -> None:
        r"""Initialise the HAT network capacity metric. Add state variables."""
        BaseAggregator.__init__(
            self,
            "sum",
            torch.tensor(0.0, dtype=torch.get_default_dtype()),
            nan_strategy,
            state_name="sum_adjustment_rate",
            **kwargs,
        )
        self.sum_adjustment_rate: Tensor
        r"""State variable created by `add_state()` to store the sum of the adjustment rate values till this layer."""

        self.add_state("num_params", default=torch.tensor(0), dist_reduce_fx="sum")
        self.num_params: Tensor
        r"""State variable created by `add_state()` to store the number of the parameters till this layer."""

    def update(
        self, adjustment_rate_weight: Tensor, adjustment_rate_bias: Tensor
    ) -> None:
        r"""Update and accumulate the sum of adjustment rate values till this layer from the layer.

        **Args:**
        - **adjustment_rate_weight** (`Tensor`): the adjustment rate values of the weight matrix of the layer.
        - **adjustment_rate_bias** (`Tensor`): the adjustment rate values of the bias vector of the layer.
        """
        adjustment_rate_weight = torch.as_tensor(
            adjustment_rate_weight, dtype=self.dtype, device=self.device
        )
        adjustment_rate_bias = torch.as_tensor(
            adjustment_rate_bias, dtype=self.dtype, device=self.device
        )

        self.sum_adjustment_rate += (
            adjustment_rate_weight.sum() + adjustment_rate_bias.sum()
        )
        self.num_params += adjustment_rate_weight.numel() + adjustment_rate_bias.numel()

    def compute(self) -> Tensor:
        r"""Compute this HAT network capacity till this layer.

        **Returns:**
        - **network_capacity** (`Tensor`): the calculated network capacity result.
        """

        return self.sum_adjustment_rate / self.num_params



r"""

# Continual Learning Algorithms

This submodule provides the **continual learning algorithms** in CLArena. 

Please note that this is an API documantation. Please refer to the main documentation pages for more information about the backbone networks and how to configure and implement them:

- [**Configure CL Algorithm**](https://pengxiang-wang.com/projects/continual-learning-arena/docs/configure-your-experiment/cl-algorithm)
- [**Implement Your CL Algorithm Class**](https://pengxiang-wang.com/projects/continual-learning-arena/docs/implement-your-cl-modules/cl-algorithm)
- [**A Beginners' Guide to Continual Learning (Methodology Overview)**](https://pengxiang-wang.com/posts/continual-learning-beginners-guide#sec-methodology)


The algorithms are implemented as subclasses of `CLAlgorithm`.

"""

from .base import CLAlgorithm

# finetuning first
from .finetuning import Finetuning
from .fix import Fix

from .lwf import LwF
from .ewc import EWC
from .cbp import CBP

from .hat import HAT
from .adahat import AdaHAT
from .cbphat import CBPHAT


__all__ = [
    "CLAlgorithm",
    "regularisers",
    "finetuning",
    "fix",
    "lwf",
    "ewc",
    "hat",
    "cbp",
    "adahat",
    "cbphat",
]



"""The module to test full experiments."""

import os

import pytest
from hydra import compose, initialize

from clarena.base import CLExperiment

experiment_list = [
    os.path.splitext(f)[0]
    for f in os.listdir("example_configs/experiment")
    if os.path.isfile(os.path.join("example_configs/experiment", f))
]


@pytest.mark.parametrize(
    "experiment",
    experiment_list,
)
def test_experiments(experiment: str):
    """Test all experiments in the experiment configs directory."""
    initialize(version_base="1.3", config_path="example_configs/")

    cfg = compose(
        config_name="entrance",
        overrides=[f"experiment={experiment}"],
        return_hydra_config=True,
    )

    # faster necessary testing
    cfg.trainer.max_epochs = 1
    cfg.num_tasks = 2

    expr = CLExperiment(cfg)
    expr.fit()



"""Main module containing entrance to run the clarena package."""

import argparse
import os
from argparse import Namespace

import hydra
from omegaconf import DictConfig

from clarena.base import (
    CLExperiment,  # It was supposed to be `from clarena import CLExperiment`, but it is not working when pip install -e . is used.
)
from clarena.utils import preprocess_config


def add_and_parse_args() -> tuple[Namespace, list[str]]:
    r"""Add and parse command line arguments that are not handled by Hydra.

    **Returns:**
    - **known_args** (`argparse.Namespace`): parsed arguments, including `--config-dir` and `--entrance`.
    - **unknown_args** (`list[str]`): unknown arguments that are handled by Hydra, for configuration overrides.
    """
    parser = argparse.ArgumentParser(
        usage="clrun [-h] [--config-dir CONFIG_DIR] [--entrance ENTRANCE] -- [experiment=EXPERIMENT_NAME] [(overrides)...]",
        description="Run a continual learning experiment. ",
        epilog="After them, you must specifiy your experiment through [experiment=EXPERIMENT_NAME]. The EXPERIMENT_NAME should be the name of the experiment YAML file in your configs/experiment directory. You can also add other overrides to the experiment configuration.",
    )
    parser.add_argument(
        "--config-dir",
        type=str,
        default="configs/",
        help="The directory path to your configs. Default: 'configs/'.",
    )
    parser.add_argument(
        "--entrance",
        type=str,
        default="entrance.yaml",
        help="The entrance YAML file name in your configs. Default: 'entrance.yaml'.",
    )
    return (
        parser.parse_known_args()
    )  # use parse_known_args() instead of parse_args() to enable Hydra overrides


def clrun() -> None:
    r"""The main entrance for running a continual learning experiment."""

    # parse the arguments
    args, _ = add_and_parse_args()

    # get the absolute path of the config directory
    hydra_config_dir = os.path.abspath(args.config_dir)

    # get the entrance YAML file name
    hydra_config_name = args.entrance

    def main(cfg: DictConfig) -> None:
        r"""A placeholder for the main codes. This function should be decorated by `hydra.main()`.

        **Args:**
        - **cfg** (`DictConfig`): the configuration for the experiment.
        """

        # preprocess the configuration before constructing the experiment
        preprocess_config(cfg)

        # construct the experiment
        expr = CLExperiment(cfg)

        # execute the experiment
        expr.run()

    # Use the specified config path for Hydra config
    hydra_decorated_main = hydra.main(
        version_base="1.3", config_path=hydra_config_dir, config_name=hydra_config_name
    )(main)

    hydra_decorated_main()
